import unittest
from .TransportLayerBaseTest import TransportLayerBaseTest
from . import unittest_logging
import isotp

Message = isotp.CanMessage

# We check that addressing modes have the right effect on payloads.
class TestAddressingMode(TransportLayerBaseTest):
    def setUp(self):
        self.init_test_case()

    def test_create_address(self):
        isotp.Address(isotp.AddressingMode.Normal_11bits, txid=1, rxid=2)
        isotp.Address(isotp.AddressingMode.Normal_29bits, txid=1, rxid=2)
        isotp.Address(isotp.AddressingMode.NormalFixed_29bits, source_address=1, target_address=2)
        isotp.Address(isotp.AddressingMode.Extended_11bits, txid=1, rxid=2, target_address=3)
        isotp.Address(isotp.AddressingMode.Extended_29bits, txid=1, rxid=2, target_address=3)
        isotp.Address(isotp.AddressingMode.Mixed_11bits, txid=1, rxid=2, address_extension=3)
        isotp.Address(isotp.AddressingMode.Mixed_29bits, source_address=1, target_address=2, address_extension=3)

    def test_single_frame_only_function_tatype(self):
        tatype = isotp.TargetAddressType.Functional

        address = isotp.Address(isotp.AddressingMode.Normal_11bits, txid=1, rxid=2)
        layer = isotp.TransportLayer(txfn=self.stack_txfn, rxfn=self.stack_rxfn, address=address)
        layer.send(self.make_payload(7), tatype)
        with self.assertRaises(ValueError):
            layer.send(self.make_payload(8), tatype)

        address = isotp.Address(isotp.AddressingMode.Normal_29bits, txid=1, rxid=2)
        layer = isotp.TransportLayer(txfn=self.stack_txfn, rxfn=self.stack_rxfn, address=address)
        layer.send(self.make_payload(7), tatype)
        with self.assertRaises(ValueError):
            layer.send(self.make_payload(8), tatype)

        address = isotp.Address(isotp.AddressingMode.NormalFixed_29bits, source_address=1, target_address=2)
        layer = isotp.TransportLayer(txfn=self.stack_txfn, rxfn=self.stack_rxfn, address=address)
        layer.send(self.make_payload(7), tatype)
        with self.assertRaises(ValueError):
            layer.send(self.make_payload(8), tatype)

        address = isotp.Address(isotp.AddressingMode.Extended_11bits, txid=1, rxid=2, target_address=3)
        layer = isotp.TransportLayer(txfn=self.stack_txfn, rxfn=self.stack_rxfn, address=address)
        layer.send(self.make_payload(6), tatype)
        with self.assertRaises(ValueError):
            layer.send(self.make_payload(7), tatype)

        address = isotp.Address(isotp.AddressingMode.Extended_29bits, txid=1, rxid=2, target_address=3)
        layer = isotp.TransportLayer(txfn=self.stack_txfn, rxfn=self.stack_rxfn, address=address)
        layer.send(self.make_payload(6), tatype)
        with self.assertRaises(ValueError):
            layer.send(self.make_payload(7), tatype)

        address = isotp.Address(isotp.AddressingMode.Mixed_11bits, txid=1, rxid=2, address_extension=3)
        layer = isotp.TransportLayer(txfn=self.stack_txfn, rxfn=self.stack_rxfn, address=address)
        layer.send(self.make_payload(6), tatype)
        with self.assertRaises(ValueError):
            layer.send(self.make_payload(7), tatype)

        address = isotp.Address(isotp.AddressingMode.Mixed_29bits, source_address=1, target_address=2, address_extension=3)
        layer = isotp.TransportLayer(txfn=self.stack_txfn, rxfn=self.stack_rxfn, address=address)
        layer.send(self.make_payload(6), tatype)
        with self.assertRaises(ValueError):
            layer.send(self.make_payload(7), tatype)
        
    def test_11bits_normal_basic(self):
        rxid = 0x123
        txid = 0x456
        address = isotp.Address(isotp.AddressingMode.Normal_11bits, txid=txid, rxid=rxid)

        self.assertTrue(address.is_for_me(Message(arbitration_id=rxid)))
        self.assertFalse(address.is_for_me(Message(arbitration_id=rxid, extended_id=True)))
        self.assertFalse(address.is_for_me(Message(arbitration_id=rxid+1)))

        self.assertEqual(address.get_tx_arbitraton_id(isotp.TargetAddressType.Physical), txid)
        self.assertEqual(address.get_tx_arbitraton_id(isotp.TargetAddressType.Functional), txid)
        self.assertEqual(address.get_rx_arbitraton_id(isotp.TargetAddressType.Physical), rxid)
        self.assertEqual(address.get_rx_arbitraton_id(isotp.TargetAddressType.Functional), rxid)

    def test_11bits_normal_through_layer(self):
        functional = isotp.TargetAddressType.Functional 
        physical = isotp.TargetAddressType.Physical 
        rxid = 0x123
        txid = 0x456
        address = isotp.Address(isotp.AddressingMode.Normal_11bits, txid=txid, rxid=rxid)
        layer = isotp.TransportLayer(txfn=self.stack_txfn, rxfn=self.stack_rxfn, address=address, params={'stmin':0, 'blocksize':0})

        # Receive Single frame - Physical
        self.simulate_rx_msg(Message(arbitration_id = rxid, data=bytearray([0x03, 0x01, 0x02, 0x03]), extended_id=False))
        layer.process()
        frame = layer.recv()
        self.assertIsNotNone(frame)
        self.assertEqual(frame, b'\x01\x02\x03')

        # Receive Single frame - Functional
        layer.reset()
        self.simulate_rx_msg(Message(arbitration_id = rxid, data=bytearray([0x03, 0x01, 0x02, 0x03]), extended_id=False))
        layer.process()
        frame = layer.recv()
        self.assertIsNotNone(frame)
        self.assertEqual(frame, b'\x01\x02\x03')

        # Receive multiframe - Physical
        layer.reset()
        self.simulate_rx_msg(Message(arbitration_id = rxid, data=bytearray([0x10, 0x08, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06]), extended_id=False))
        layer.process()
        self.assert_sent_flow_control(stmin=0, blocksize=0)
        self.simulate_rx_msg(Message(arbitration_id = rxid, data=bytearray([0x21, 0x07, 0x08]), extended_id=False))
        layer.process()
        frame = layer.recv()
        self.assertIsNotNone(frame)
        self.assertEqual(frame, b'\x01\x02\x03\x04\x05\x06\x07\x08')

        #Transmit single frame - Physical / Functional
        layer.reset()
        layer.send(b'\x04\x05\x06', physical)
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid)
        self.assertEqual(msg.data, bytearray([0x03, 0x04, 0x05, 0x06]))
        self.assertFalse(msg.is_extended_id)

        # Transmit multiframe - Physical
        layer.reset()
        layer.send(b'\x04\x05\x06\x07\x08\x09\x0A\x0B', physical)
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid)
        self.assertEqual(msg.data, bytearray([0x10, 0x08, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09]))
        self.assertFalse(msg.is_extended_id)

        self.simulate_rx_msg(Message(arbitration_id=rxid, data=self.make_flow_control_data(flow_status=0, stmin=0, blocksize=0), extended_id=False))
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid)
        self.assertEqual(msg.data, bytearray([0x21, 0x0A, 0x0B]))
        self.assertFalse(msg.is_extended_id)

    def test_29bits_normal_basic(self):
        rxid = 0x123456
        txid = 0x789ABC

        address = isotp.Address(isotp.AddressingMode.Normal_29bits, txid=txid, rxid=rxid)
        self.assertTrue(address.is_for_me(Message(arbitration_id=rxid,  extended_id=True)))
        self.assertFalse(address.is_for_me(Message(arbitration_id=rxid, extended_id=False)))
        self.assertFalse(address.is_for_me(Message(arbitration_id=rxid+1, extended_id=True)))
        self.assertFalse(address.is_for_me(Message(arbitration_id=rxid+1, extended_id=False)))

        self.assertEqual(address.get_tx_arbitraton_id(isotp.TargetAddressType.Physical), txid)
        self.assertEqual(address.get_tx_arbitraton_id(isotp.TargetAddressType.Functional), txid)
        self.assertEqual(address.get_rx_arbitraton_id(isotp.TargetAddressType.Physical), rxid)
        self.assertEqual(address.get_rx_arbitraton_id(isotp.TargetAddressType.Functional), rxid)

    def test_29bits_normal_through_layer(self):
        functional = isotp.TargetAddressType.Functional 
        physical = isotp.TargetAddressType.Physical 
        rxid = 0x123456
        txid = 0x789ABC
        address = isotp.Address(isotp.AddressingMode.Normal_29bits, txid=txid, rxid=rxid)
        layer = isotp.TransportLayer(txfn=self.stack_txfn, rxfn=self.stack_rxfn, address=address, params={'stmin':0, 'blocksize':0})

        # Receive Single frame - Physical
        self.simulate_rx_msg(Message(arbitration_id = rxid, data=bytearray([0x03, 0x01, 0x02, 0x03]), extended_id=True))
        layer.process()
        frame = layer.recv()
        self.assertIsNotNone(frame)
        self.assertEqual(frame, b'\x01\x02\x03')

        # Receive Single frame - Functional
        layer.reset()
        self.simulate_rx_msg(Message(arbitration_id = rxid, data=bytearray([0x03, 0x01, 0x02, 0x03]), extended_id=True))
        layer.process()
        frame = layer.recv()
        self.assertIsNotNone(frame)
        self.assertEqual(frame, b'\x01\x02\x03')

        # Receive multiframe - Physical
        layer.reset()
        self.simulate_rx_msg(Message(arbitration_id = rxid, data=bytearray([0x10, 0x08, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06]), extended_id=True))
        layer.process()
        self.assert_sent_flow_control(stmin=0, blocksize=0)
        self.simulate_rx_msg(Message(arbitration_id = rxid, data=bytearray([0x21, 0x07, 0x08]), extended_id=True))
        layer.process()
        frame = layer.recv()
        self.assertIsNotNone(frame)
        self.assertEqual(frame, b'\x01\x02\x03\x04\x05\x06\x07\x08')

        #Transmit single frame - Physical / Functional
        layer.reset()
        layer.send(b'\x04\x05\x06', physical)
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid)
        self.assertEqual(msg.data, bytearray([0x03, 0x04, 0x05, 0x06]))
        self.assertTrue(msg.is_extended_id)

        # Transmit multiframe - Physical
        layer.reset()
        layer.send(b'\x04\x05\x06\x07\x08\x09\x0A\x0B', physical)
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid)
        self.assertEqual(msg.data, bytearray([0x10, 0x08, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09]))
        self.assertTrue(msg.is_extended_id)

        self.simulate_rx_msg(Message(arbitration_id=rxid, data=self.make_flow_control_data(flow_status=0, stmin=0, blocksize=0), extended_id=True))
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid)
        self.assertEqual(msg.data, bytearray([0x21, 0x0A, 0x0B]))
        self.assertTrue(msg.is_extended_id)

    def test_29bits_normal_fixed(self):
        ta = 0x55
        sa = 0xAA
        rxid_physical = 0x18DAAA55
        rxid_functional = 0x18DBAA55
        txid_physical = 0x18DA55AA
        txid_functional = 0x18DB55AA
        
        address = isotp.Address(isotp.AddressingMode.NormalFixed_29bits, target_address = ta, source_address=sa)

        self.assertTrue(address.is_for_me(Message(rxid_physical,  extended_id=True)))
        self.assertTrue(address.is_for_me(Message(rxid_functional,  extended_id=True)))
        self.assertFalse(address.is_for_me(Message(txid_physical,  extended_id=True)))
        self.assertFalse(address.is_for_me(Message(txid_functional,  extended_id=True)))
        self.assertFalse(address.is_for_me(Message(arbitration_id=(rxid_physical) & 0x7FF, extended_id=False)))
        self.assertFalse(address.is_for_me(Message(arbitration_id=rxid_physical+1, extended_id=True)))
        self.assertFalse(address.is_for_me(Message(arbitration_id=(rxid_physical+1)&0x7FF, extended_id=False)))

        self.assertEqual(address.get_tx_arbitraton_id(isotp.TargetAddressType.Physical), txid_physical)
        self.assertEqual(address.get_tx_arbitraton_id(isotp.TargetAddressType.Functional), txid_functional)
        self.assertEqual(address.get_rx_arbitraton_id(isotp.TargetAddressType.Physical), rxid_physical)
        self.assertEqual(address.get_rx_arbitraton_id(isotp.TargetAddressType.Functional), rxid_functional)

    def test_29bits_normal_fixed_through_layer(self):
        functional = isotp.TargetAddressType.Functional 
        physical = isotp.TargetAddressType.Physical 
        ta = 0x55
        sa = 0xAA
        rxid_physical = 0x18DAAA55
        rxid_functional = 0x18DBAA55
        txid_physical = 0x18DA55AA
        txid_functional = 0x18DB55AA
        
        address = isotp.Address(isotp.AddressingMode.NormalFixed_29bits, target_address = ta, source_address=sa)
        layer = isotp.TransportLayer(txfn=self.stack_txfn, rxfn=self.stack_rxfn, address=address, params={'stmin':0, 'blocksize':0})

        # Receive Single frame - Physical
        self.simulate_rx_msg(Message(arbitration_id = rxid_physical, data=bytearray([0x03, 0x01, 0x02, 0x03]), extended_id=True))
        layer.process()
        frame = layer.recv()
        self.assertIsNotNone(frame)
        self.assertEqual(frame, b'\x01\x02\x03')

        # Receive Single frame - Functional
        layer.reset()
        self.simulate_rx_msg(Message(arbitration_id = rxid_functional, data=bytearray([0x03, 0x01, 0x02, 0x03]), extended_id=True))
        layer.process()
        frame = layer.recv()
        self.assertIsNotNone(frame)
        self.assertEqual(frame, b'\x01\x02\x03')

        # Receive multiframe - Physical
        layer.reset()
        self.simulate_rx_msg(Message(arbitration_id = rxid_physical, data=bytearray([0x10, 0x08, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06]), extended_id=True))
        layer.process()
        self.assert_sent_flow_control(stmin=0, blocksize=0)
        self.simulate_rx_msg(Message(arbitration_id = rxid_physical, data=bytearray([0x21, 0x07, 0x08]), extended_id=True))
        layer.process()
        frame = layer.recv()
        self.assertIsNotNone(frame)
        self.assertEqual(frame, b'\x01\x02\x03\x04\x05\x06\x07\x08')

        #Transmit single frame - Physical
        layer.reset()
        layer.send(b'\x04\x05\x06', physical)
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid_physical)
        self.assertEqual(msg.data, bytearray([0x03, 0x04, 0x05, 0x06]))
        self.assertTrue(msg.is_extended_id)

        #Transmit single frame - Functional
        layer.reset()
        layer.send(b'\x04\x05\x06', functional)
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid_functional)
        self.assertEqual(msg.data, bytearray([0x03, 0x04, 0x05, 0x06]))
        self.assertTrue(msg.is_extended_id)

        # Transmit multiframe - Physical
        layer.reset()
        layer.send(b'\x04\x05\x06\x07\x08\x09\x0A\x0B', physical)
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid_physical)
        self.assertEqual(msg.data, bytearray([0x10, 0x08, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09]))
        self.assertTrue(msg.is_extended_id)

        self.simulate_rx_msg(Message(arbitration_id=rxid_physical, data=self.make_flow_control_data(flow_status=0, stmin=0, blocksize=0), extended_id=True))
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid_physical)
        self.assertEqual(msg.data, bytearray([0x21, 0x0A, 0x0B]))
        self.assertTrue(msg.is_extended_id)

    def test_11bits_extended(self):
        txid = 0x123
        rxid = 0x456
        sa = 0x55
        ta = 0xAA
        
        address = isotp.Address(isotp.AddressingMode.Extended_11bits, txid=txid, rxid=rxid, source_address=sa, target_address=ta)

        self.assertFalse(address.is_for_me(Message(rxid,  extended_id=False))) # No data
        self.assertFalse(address.is_for_me(Message(txid,  extended_id=False))) # No data, wrong id
        self.assertFalse(address.is_for_me(Message(rxid, data = bytearray([ta]),  extended_id=False))) # wrong id
        self.assertTrue(address.is_for_me(Message(rxid, data = bytearray([sa]),  extended_id=False)))
        self.assertFalse(address.is_for_me(Message(rxid, data = bytearray([sa]),  extended_id=True)))
        self.assertFalse(address.is_for_me(Message(rxid+1, data = bytearray([sa]),  extended_id=False)))
        self.assertFalse(address.is_for_me(Message(rxid, data = bytearray([sa+1]),  extended_id=False)))

        self.assertEqual(address.get_tx_arbitraton_id(isotp.TargetAddressType.Physical), txid)
        self.assertEqual(address.get_tx_arbitraton_id(isotp.TargetAddressType.Functional), txid)
        self.assertEqual(address.get_rx_arbitraton_id(isotp.TargetAddressType.Physical), rxid)
        self.assertEqual(address.get_rx_arbitraton_id(isotp.TargetAddressType.Functional), rxid)

    def test_11bits_extended_through_layer(self):
        functional = isotp.TargetAddressType.Functional 
        physical = isotp.TargetAddressType.Physical 
        txid = 0x123
        rxid = 0x456
        sa = 0x55
        ta = 0xAA
        
        address = isotp.Address(isotp.AddressingMode.Extended_11bits, txid=txid, rxid=rxid, source_address=sa, target_address=ta)
        layer = isotp.TransportLayer(txfn=self.stack_txfn, rxfn=self.stack_rxfn, address=address, params={'stmin':0, 'blocksize':0})

        # Receive Single frame - Physical / Functional
        self.simulate_rx_msg(Message(arbitration_id = rxid, data=bytearray([sa, 0x03, 0x01, 0x02, 0x03]), extended_id=False))
        layer.process()
        frame = layer.recv()
        self.assertIsNotNone(frame)
        self.assertEqual(frame, b'\x01\x02\x03')

        # Receive multiframe - Physical
        layer.reset()
        self.simulate_rx_msg(Message(arbitration_id = rxid, data=bytearray([sa, 0x10, 0x08, 0x01, 0x02, 0x03, 0x04, 0x05]), extended_id=False))
        layer.process()
        self.assert_sent_flow_control(prefix=[ta], stmin=0, blocksize=0)
        self.simulate_rx_msg(Message(arbitration_id = rxid, data=bytearray([sa, 0x21, 0x06, 0x07, 0x08]), extended_id=False))
        layer.process()
        frame = layer.recv()
        self.assertIsNotNone(frame)
        self.assertEqual(frame, b'\x01\x02\x03\x04\x05\x06\x07\x08')


        #Transmit single frame - Physical
        layer.reset()
        layer.send(b'\x04\x05\x06', physical)
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid)
        self.assertEqual(msg.data, bytearray([ta, 0x03, 0x04, 0x05, 0x06]))
        self.assertFalse(msg.is_extended_id)

        #Transmit single frame - Functional
        layer.reset()
        layer.send(b'\x04\x05\x06', functional)
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid)
        self.assertEqual(msg.data, bytearray([ta, 0x03, 0x04, 0x05, 0x06]))
        self.assertFalse(msg.is_extended_id)

        # Transmit multiframe - Physical
        layer.reset()
        layer.send(b'\x04\x05\x06\x07\x08\x09\x0A\x0B', physical)
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid)
        self.assertEqual(msg.data, bytearray([ta, 0x10, 0x08, 0x04, 0x05, 0x06, 0x07, 0x08]))
        self.assertFalse(msg.is_extended_id)

        self.simulate_rx_msg(Message(arbitration_id=rxid, data=self.make_flow_control_data(flow_status=0, stmin=0, blocksize=0, prefix=[sa]), extended_id=False))
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid)
        self.assertEqual(msg.data, bytearray([ta, 0x21, 0x09, 0x0A, 0x0B]))
        self.assertFalse(msg.is_extended_id)

    def test_29bits_extended(self):
        txid = 0x123
        rxid = 0x456
        sa = 0x55
        ta = 0xAA
        
        address = isotp.Address(isotp.AddressingMode.Extended_29bits, txid=txid, rxid=rxid, source_address=sa, target_address=ta)

        self.assertFalse(address.is_for_me(Message(rxid,  extended_id=True))) # No data
        self.assertFalse(address.is_for_me(Message(txid,  extended_id=True))) # No data, wrong id
        self.assertFalse(address.is_for_me(Message(rxid, data = bytearray([ta]),  extended_id=True))) # wrong id
        self.assertTrue(address.is_for_me(Message(rxid, data = bytearray([sa]),  extended_id=True)))
        self.assertFalse(address.is_for_me(Message(rxid, data = bytearray([sa]),  extended_id=False)))
        self.assertFalse(address.is_for_me(Message(rxid+1, data = bytearray([sa]),  extended_id=True)))
        self.assertFalse(address.is_for_me(Message(rxid, data = bytearray([sa+1]),  extended_id=True)))

        self.assertEqual(address.get_tx_arbitraton_id(isotp.TargetAddressType.Physical ), txid)
        self.assertEqual(address.get_tx_arbitraton_id(isotp.TargetAddressType.Functional), txid)
        self.assertEqual(address.get_rx_arbitraton_id(isotp.TargetAddressType.Physical), rxid)
        self.assertEqual(address.get_rx_arbitraton_id(isotp.TargetAddressType.Functional), rxid)

    def test_29bits_extended_through_layer(self):
        functional = isotp.TargetAddressType.Functional 
        physical = isotp.TargetAddressType.Physical 
        txid = 0x123
        rxid = 0x456
        sa = 0x55
        ta = 0xAA
        
        address = isotp.Address(isotp.AddressingMode.Extended_29bits, txid=txid, rxid=rxid, source_address=sa, target_address=ta)
        layer = isotp.TransportLayer(txfn=self.stack_txfn, rxfn=self.stack_rxfn, address=address, params={'stmin':0, 'blocksize':0})

        # Receive Single frame - Physical / Functional
        self.simulate_rx_msg(Message(arbitration_id = rxid, data=bytearray([sa, 0x03, 0x01, 0x02, 0x03]), extended_id=True))
        layer.process()
        frame = layer.recv()
        self.assertIsNotNone(frame)
        self.assertEqual(frame, b'\x01\x02\x03')

        # Receive multiframe - Physical
        layer.reset()
        self.simulate_rx_msg(Message(arbitration_id = rxid, data=bytearray([sa, 0x10, 0x08, 0x01, 0x02, 0x03, 0x04, 0x05]), extended_id=True))
        layer.process()
        self.assert_sent_flow_control(prefix=[ta], stmin=0, blocksize=0)
        self.simulate_rx_msg(Message(arbitration_id = rxid, data=bytearray([sa, 0x21, 0x06, 0x07, 0x08]), extended_id=True))
        layer.process()
        frame = layer.recv()
        self.assertIsNotNone(frame)
        self.assertEqual(frame, b'\x01\x02\x03\x04\x05\x06\x07\x08')


        #Transmit single frame - Physical
        layer.reset()
        layer.send(b'\x04\x05\x06', physical)
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid)
        self.assertEqual(msg.data, bytearray([ta, 0x03, 0x04, 0x05, 0x06]))
        self.assertTrue(msg.is_extended_id)

        #Transmit single frame - Functional
        layer.reset()
        layer.send(b'\x04\x05\x06', functional)
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid)
        self.assertEqual(msg.data, bytearray([ta, 0x03, 0x04, 0x05, 0x06]))
        self.assertTrue(msg.is_extended_id)

        # Transmit multiframe - Physical
        layer.reset()
        layer.send(b'\x04\x05\x06\x07\x08\x09\x0A\x0B', physical)
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid)
        self.assertEqual(msg.data, bytearray([ta, 0x10, 0x08, 0x04, 0x05, 0x06, 0x07, 0x08]))
        self.assertTrue(msg.is_extended_id)

        self.simulate_rx_msg(Message(arbitration_id=rxid, data=self.make_flow_control_data(flow_status=0, stmin=0, blocksize=0, prefix=[sa]), extended_id=True))
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid)
        self.assertEqual(msg.data, bytearray([ta, 0x21, 0x09, 0x0A, 0x0B]))
        self.assertTrue(msg.is_extended_id)

    def test_11bits_mixed(self):
        txid = 0x123
        rxid = 0x456
        ae = 0x99
        
        address = isotp.Address(isotp.AddressingMode.Mixed_11bits, txid=txid, rxid=rxid, address_extension = ae)

        self.assertFalse(address.is_for_me(Message(rxid,  extended_id=False))) # No data
        self.assertFalse(address.is_for_me(Message(txid,  extended_id=False))) # No data, wrong id
        self.assertTrue(address.is_for_me(Message(rxid, data = bytearray([ae]),  extended_id=False)))
        self.assertFalse(address.is_for_me(Message(rxid, data = bytearray([ae]),  extended_id=True)))
        self.assertFalse(address.is_for_me(Message(rxid+1, data = bytearray([ae]),  extended_id=False)))
        self.assertFalse(address.is_for_me(Message(rxid, data = bytearray([ae+1]),  extended_id=False)))

        self.assertEqual(address.get_tx_arbitraton_id(isotp.TargetAddressType.Physical), txid)
        self.assertEqual(address.get_tx_arbitraton_id(isotp.TargetAddressType.Functional), txid)
        self.assertEqual(address.get_rx_arbitraton_id(isotp.TargetAddressType.Physical), rxid)
        self.assertEqual(address.get_rx_arbitraton_id(isotp.TargetAddressType.Functional), rxid)

    def test_11bits_mixed_through_layer(self):
        functional = isotp.TargetAddressType.Functional 
        physical = isotp.TargetAddressType.Physical 
        txid = 0x123
        rxid = 0x456
        ae = 0x99
        
        address = isotp.Address(isotp.AddressingMode.Mixed_11bits, txid=txid, rxid=rxid, address_extension = ae)
        layer = isotp.TransportLayer(txfn=self.stack_txfn, rxfn=self.stack_rxfn, address=address, params={'stmin':0, 'blocksize':0})

        # Receive Single frame - Physical / Functional
        self.simulate_rx_msg(Message(arbitration_id = rxid, data=bytearray([ae, 0x03, 0x01, 0x02, 0x03]), extended_id=False))
        layer.process()
        frame = layer.recv()
        self.assertIsNotNone(frame)
        self.assertEqual(frame, b'\x01\x02\x03')

        # Receive multiframe - Physical
        layer.reset()
        self.simulate_rx_msg(Message(arbitration_id = rxid, data=bytearray([ae, 0x10, 0x08, 0x01, 0x02, 0x03, 0x04, 0x05]), extended_id=False))
        layer.process()
        self.assert_sent_flow_control(prefix=[ae], stmin=0, blocksize=0)
        self.simulate_rx_msg(Message(arbitration_id = rxid, data=bytearray([ae, 0x21, 0x06, 0x07, 0x08]), extended_id=False))
        layer.process()
        frame = layer.recv()
        self.assertIsNotNone(frame)
        self.assertEqual(frame, b'\x01\x02\x03\x04\x05\x06\x07\x08')


        #Transmit single frame - Physical
        layer.reset()
        layer.send(b'\x04\x05\x06', physical)
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid)
        self.assertEqual(msg.data, bytearray([ae, 0x03, 0x04, 0x05, 0x06]))
        self.assertFalse(msg.is_extended_id)

        #Transmit single frame - Functional
        layer.reset()
        layer.send(b'\x04\x05\x06', functional)
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid)
        self.assertEqual(msg.data, bytearray([ae, 0x03, 0x04, 0x05, 0x06]))
        self.assertFalse(msg.is_extended_id)

        # Transmit multiframe - Physical
        layer.reset()
        layer.send(b'\x04\x05\x06\x07\x08\x09\x0A\x0B', physical)
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid)
        self.assertEqual(msg.data, bytearray([ae, 0x10, 0x08, 0x04, 0x05, 0x06, 0x07, 0x08]))
        self.assertFalse(msg.is_extended_id)

        self.simulate_rx_msg(Message(arbitration_id=rxid, data=self.make_flow_control_data(flow_status=0, stmin=0, blocksize=0, prefix=[ae]), extended_id=False))
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid)
        self.assertEqual(msg.data, bytearray([ae, 0x21, 0x09, 0x0A, 0x0B]))
        self.assertFalse(msg.is_extended_id)

    def test_29bits_mixed(self):
        ta = 0x55
        sa = 0xAA
        ae = 0x99
        rxid_physical   = 0x18CEAA55
        rxid_functional = 0x18CDAA55
        txid_physical   = 0x18CE55AA
        txid_functional = 0x18CD55AA

        address = isotp.Address(isotp.AddressingMode.Mixed_29bits, source_address=sa, target_address=ta, address_extension = ae)

        self.assertFalse(address.is_for_me(Message(rxid_physical,   extended_id=True)))     # No data
        self.assertFalse(address.is_for_me(Message(rxid_functional, extended_id=True)))     # No data
        self.assertFalse(address.is_for_me(Message(txid_physical,   extended_id=True)))     # No data
        self.assertFalse(address.is_for_me(Message(txid_functional, extended_id=True)))     # No data

        self.assertTrue(address.is_for_me(Message(rxid_physical,    data = bytearray([ae]),  extended_id=True)))
        self.assertFalse(address.is_for_me(Message(rxid_physical,   data = bytearray([ae]),  extended_id=False)))
        self.assertTrue(address.is_for_me(Message(rxid_functional,  data = bytearray([ae]),  extended_id=True)))
        self.assertFalse(address.is_for_me(Message(rxid_functional,  data = bytearray([ae]),  extended_id=False)))
        self.assertFalse(address.is_for_me(Message(txid_physical,   data = bytearray([ae]),  extended_id=True)))
        self.assertFalse(address.is_for_me(Message(txid_functional, data = bytearray([ae]),  extended_id=True)))

        self.assertEqual(address.get_tx_arbitraton_id(isotp.TargetAddressType.Physical),    txid_physical)
        self.assertEqual(address.get_tx_arbitraton_id(isotp.TargetAddressType.Functional),  txid_functional)
        self.assertEqual(address.get_rx_arbitraton_id(isotp.TargetAddressType.Physical),    rxid_physical)
        self.assertEqual(address.get_rx_arbitraton_id(isotp.TargetAddressType.Functional),  rxid_functional)

    def test_29bits_mixed_through_layer(self):
        functional = isotp.TargetAddressType.Functional 
        physical = isotp.TargetAddressType.Physical 
        ta = 0x55
        sa = 0xAA
        ae = 0x99
        rxid_physical = 0x18CEAA55
        rxid_functional = 0x18CDAA55
        txid_physical = 0x18CE55AA
        txid_functional = 0x18CD55AA
        
        address = isotp.Address(isotp.AddressingMode.Mixed_29bits, source_address=sa, target_address=ta, address_extension = ae)
        layer = isotp.TransportLayer(txfn=self.stack_txfn, rxfn=self.stack_rxfn, address=address, params={'stmin':0, 'blocksize':0})

        # Receive Single frame - Physical
        self.simulate_rx_msg(Message(arbitration_id = rxid_physical, data=bytearray([ae, 0x03, 0x01, 0x02, 0x03]), extended_id=True))
        layer.process()
        frame = layer.recv()
        self.assertIsNotNone(frame)
        self.assertEqual(frame, b'\x01\x02\x03')

        # Receive Single frame - Functional
        layer.reset()
        self.simulate_rx_msg(Message(arbitration_id = rxid_functional, data=bytearray([ae, 0x03, 0x01, 0x02, 0x03]), extended_id=True))
        layer.process()
        frame = layer.recv()
        self.assertIsNotNone(frame)
        self.assertEqual(frame, b'\x01\x02\x03')

        # Receive multiframe - Physical
        layer.reset()
        self.simulate_rx_msg(Message(arbitration_id = rxid_physical, data=bytearray([ae, 0x10, 0x08, 0x01, 0x02, 0x03, 0x04, 0x05]), extended_id=True))
        layer.process()
        self.assert_sent_flow_control(prefix=[ae], stmin=0, blocksize=0)
        self.simulate_rx_msg(Message(arbitration_id = rxid_physical, data=bytearray([ae, 0x21, 0x06, 0x07, 0x08]), extended_id=True))
        layer.process()
        frame = layer.recv()
        self.assertIsNotNone(frame)
        self.assertEqual(frame, b'\x01\x02\x03\x04\x05\x06\x07\x08')

        #Transmit single frame - Physical
        layer.reset()
        layer.send(b'\x04\x05\x06', physical)
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid_physical)
        self.assertEqual(msg.data, bytearray([ae, 0x03, 0x04, 0x05, 0x06]))
        self.assertTrue(msg.is_extended_id)

        #Transmit single frame - Functional
        layer.reset()
        layer.send(b'\x04\x05\x06', functional)
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid_functional)
        self.assertEqual(msg.data, bytearray([ae, 0x03, 0x04, 0x05, 0x06]))
        self.assertTrue(msg.is_extended_id)

        # Transmit multiframe - Physical
        layer.reset()
        layer.send(b'\x04\x05\x06\x07\x08\x09\x0A\x0B', physical)
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid_physical)
        self.assertEqual(msg.data, bytearray([ae, 0x10, 0x08, 0x04, 0x05, 0x06, 0x07, 0x08]))
        self.assertTrue(msg.is_extended_id)

        self.simulate_rx_msg(Message(arbitration_id=rxid_physical, data=self.make_flow_control_data(flow_status=0, stmin=0, blocksize=0, prefix=[ae]), extended_id=True))
        layer.process()
        msg = self.get_tx_can_msg()
        self.assertIsNotNone(msg)
        self.assertEqual(msg.arbitration_id, txid_physical)
        self.assertEqual(msg.data, bytearray([ae, 0x21, 0x09, 0x0A, 0x0B]))
        self.assertTrue(msg.is_extended_id)



import isotp
import queue
import os
import binascii
import time
import unittest
from functools import partial
import isotp
from . import unittest_logging
from .TransportLayerBaseTest import TransportLayerBaseTest
Message = isotp.CanMessage


# Check the behaviour of the transport layer. Sequenece of CAN frames, timings, etc.
class TestTransportLayer(TransportLayerBaseTest):
	RXID = 0x456
	TXID = 0x123

	def setUp(self):
		params = {
			'stmin' : 1,
			'blocksize' : 8,
			'squash_stmin_requirement' : False,
			'rx_flowcontrol_timeout'  : 1000,
			'rx_consecutive_frame_timeout' : 1000,
			'wftmax' : 0
		}
		address = isotp.Address(isotp.AddressingMode.Normal_11bits, txid=self.TXID, rxid=self.RXID)
		self.stack = isotp.TransportLayer(txfn=self.stack_txfn, rxfn=self.stack_rxfn, address=address, error_handler = self.error_handler, params = params)
		self.error_triggered = {}

		self.init_test_case()


	def simulate_rx(self, data, rxid=RXID):
		self.simulate_rx_msg(Message(arbitration_id=rxid, data=bytearray(data)))

	def simulate_rx_flowcontrol(self, flow_status, stmin, blocksize, prefix=None):
		data = bytearray()
		if prefix is not None:
			data.extend(bytearray(prefix))
		data.extend(bytearray( [0x30 | (flow_status & 0xF), blocksize, stmin]))

		self.simulate_rx(data = data)

	# Make sure we can receive a single frame
	def test_receive_single_sf(self):
		self.simulate_rx(data = [0x05, 0x11, 0x22, 0x33, 0x44, 0x55])
		self.stack.process()
		self.assertEqual(self.rx_isotp_frame(), bytearray([0x11, 0x22, 0x33, 0x44, 0x55]))

	# Make sure we can receive multiple single frame
	def test_receive_multiple_sf(self):
		self.stack.process()
		self.stack.process()

		self.simulate_rx(data = [0x05, 0x11, 0x22, 0x33, 0x44, 0x55])
		self.stack.process()
		self.assertEqual(self.rx_isotp_frame(), bytearray([0x11, 0x22, 0x33, 0x44, 0x55])) 	

		self.assertIsNone(self.rx_isotp_frame())
		self.stack.process()
		self.assertIsNone(self.rx_isotp_frame())

		self.simulate_rx(data = [0x05, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE])
		self.stack.process()
		self.assertEqual(self.rx_isotp_frame(), bytearray([0xAA, 0xBB, 0xCC, 0xDD, 0xEE]))

		self.assertIsNone(self.rx_isotp_frame())
		self.stack.process()
		self.assertIsNone(self.rx_isotp_frame())

	def test_receive_multiple_sf_single_process_call(self):
		self.simulate_rx(data = [0x05, 0x11, 0x22, 0x33, 0x44, 0x55])
		self.simulate_rx(data = [0x05, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE])
		self.stack.process()	# Call process once
		self.assertEqual(self.rx_isotp_frame(), bytearray([0x11, 0x22, 0x33, 0x44, 0x55]))
		self.assertEqual(self.rx_isotp_frame(), bytearray([0xAA, 0xBB, 0xCC, 0xDD, 0xEE]))
		self.assertIsNone(self.rx_isotp_frame())

	def test_receive_multiframe(self):
		payload_size = 10
		payload = self.make_payload(payload_size)
		self.simulate_rx(data = [0x10, payload_size] + payload[0:6])
		self.simulate_rx(data = [0x21] + payload[6:10])
		self.stack.process()
		data = self.rx_isotp_frame()
		self.assertEqual(data, bytearray(payload))
		self.assertIsNone(self.rx_isotp_frame())

	def test_receive_2_multiframe(self):
		payload_size = 10
		payload = self.make_payload(payload_size)

		self.simulate_rx(data = [0x10, payload_size] + payload[0:6])
		self.simulate_rx(data = [0x21] + payload[6:10])
		self.simulate_rx(data = [0x10, payload_size] + payload[0:6])
		self.simulate_rx(data = [0x21] + payload[6:10])
		self.stack.process()
		self.assertEqual(self.rx_isotp_frame(), bytearray(payload))
		self.assertEqual(self.rx_isotp_frame(), bytearray(payload))
		self.assertIsNone(self.rx_isotp_frame())

	def test_receive_multiframe_check_flowcontrol(self):
		self.stack.params.set('stmin', 0x02)
		self.stack.params.set('blocksize', 0x05)

		payload_size = 10
		payload = self.make_payload(payload_size)
		self.simulate_rx(data = [0x10, payload_size] + payload[0:6])
		self.stack.process()
		self.assert_sent_flow_control(stmin=2, blocksize=5)
		self.assertIsNone(self.rx_isotp_frame())
		self.simulate_rx(data = [0x21] + payload[6:10])
		self.stack.process()
		data = self.rx_isotp_frame()
		self.assertEqual(data, bytearray(payload))
		self.assertIsNone(self.rx_isotp_frame())

	def test_receive_multiframe_flowcontrol_padding(self):
		padding_byte = 0x22
		self.stack.params.set('tx_padding', padding_byte)
		self.stack.params.set('stmin', 0x02)
		self.stack.params.set('blocksize', 0x05)

		payload_size = 10
		payload = self.make_payload(payload_size)
		self.simulate_rx(data = [0x10, payload_size] + payload[0:6])
		self.stack.process()
		self.assert_sent_flow_control(stmin=2, blocksize=5, padding_byte=padding_byte)
		self.assertIsNone(self.rx_isotp_frame())
		self.simulate_rx(data = [0x21] + payload[6:10])
		self.stack.process()
		data = self.rx_isotp_frame()
		self.assertEqual(data, bytearray(payload))
		self.assertIsNone(self.rx_isotp_frame())

	def test_long_multiframe_2_flow_control(self):
		payload_size = 30
		payload = self.make_payload(payload_size)
		self.stack.params.set('stmin', 0x05)
		self.stack.params.set('blocksize', 0x3)
		self.simulate_rx(data = [0x10, payload_size] + payload[0:6])
		self.stack.process()
		self.assert_sent_flow_control(stmin=5, blocksize=3)
		self.assertIsNone(self.rx_isotp_frame())
		self.simulate_rx(data = [0x21] + payload[6:13])
		self.stack.process()
		self.assertIsNone(self.rx_isotp_frame())
		self.simulate_rx(data = [0x22] + payload[13:20])
		self.stack.process()
		self.assertIsNone(self.rx_isotp_frame())
		self.simulate_rx(data = [0x23] + payload[20:27])
		self.stack.process()
		self.assert_sent_flow_control(stmin=5, blocksize=3)
		self.assertIsNone(self.rx_isotp_frame())
		self.simulate_rx(data = [0x24] + payload[27:30])
		self.stack.process()		
		data = self.rx_isotp_frame()
		self.assertEqual(data, bytearray(payload))
		self.assertIsNone(self.rx_isotp_frame())

	def test_receive_multiframe_bad_seqnum(self):
		payload_size = 10
		payload = self.make_payload(payload_size)
		self.simulate_rx(data = [0x10, payload_size] + payload[0:6])
		self.simulate_rx(data = [0x22] + payload[6:10])		# Bad Sequence number
		self.stack.process()
		self.assertIsNone(self.rx_isotp_frame())
		self.assertIsNone(self.get_tx_can_msg()) # Do not send flow control
		self.assert_error_triggered(isotp.WrongSequenceNumberError)

	def test_receive_timeout_consecutive_frame_after_first_frame(self):
		self.stack.params.set('rx_consecutive_frame_timeout', 200)

		payload_size = 10
		payload = self.make_payload(payload_size)
		self.simulate_rx(data = [0x10, payload_size] + payload[0:6])
		self.stack.process()
		time.sleep(0.2)	# Should stop receivving after 200 msec
		self.simulate_rx(data = [0x21] + payload[6:10])
		self.stack.process()
		self.assertIsNone(self.rx_isotp_frame())	# No message received indeed
		self.assert_error_triggered(isotp.ConsecutiveFrameTimeoutError)
		self.assert_error_triggered(isotp.UnexpectedConsecutiveFrameError)

	def test_receive_recover_timeout_consecutive_frame(self):
		self.stack.params.set('rx_consecutive_frame_timeout', 200)

		payload_size = 10
		payload1 = self.make_payload(payload_size)
		payload2 = self.make_payload(payload_size,1)
		self.assertNotEqual(payload1, payload2)
		self.simulate_rx(data = [0x10, payload_size] + payload1[0:6])
		self.stack.process()
		time.sleep(0.2)
		self.simulate_rx(data = [0x21] + payload1[6:10])
		self.stack.process()
		self.assertIsNone(self.rx_isotp_frame())
		self.simulate_rx(data = [0x10, payload_size] + payload2[0:6])
		self.simulate_rx(data = [0x21] + payload2[6:10])	
		self.stack.process()
		self.assertEqual(self.rx_isotp_frame(), bytearray(payload2))	# Correctly received
		self.assert_error_triggered(isotp.ConsecutiveFrameTimeoutError)
		self.assert_error_triggered(isotp.UnexpectedConsecutiveFrameError)

	def test_receive_multiframe_interrupting_another(self):
		self.stack.params.set('rx_consecutive_frame_timeout', 200)

		payload_size = 10
		payload1 = self.make_payload(payload_size)
		payload2 = self.make_payload(payload_size,1)
		self.assertNotEqual(payload1, payload2)
		self.simulate_rx(data = [0x10, payload_size] + payload1[0:6])
		self.simulate_rx(data = [0x10, payload_size] + payload2[0:6])	# New frame interrupting previous
		self.simulate_rx(data = [0x21] + payload2[6:10])		
		self.stack.process()
		self.assertEqual(self.rx_isotp_frame(), bytearray(payload2))
		self.assertIsNone(self.rx_isotp_frame())
		self.assert_error_triggered(isotp.ReceptionInterruptedWithFirstFrameError)

	def test_receive_single_frame_interrupt_multiframe_then_recover(self):
		self.stack.params.set('rx_consecutive_frame_timeout', 200)

		payload1_size = 16
		payload2_size = 16
		payload1 = self.make_payload(payload1_size)
		payload2 = self.make_payload(payload2_size,1)
		sf_payload = self.make_payload(5, 2)
		self.assertNotEqual(payload1, payload2)
		self.simulate_rx(data = [0x10, payload1_size] + payload1[0:6])
		self.stack.process()
		self.simulate_rx(data = [0x21] + payload1[6:13])	
		self.simulate_rx(data = [0x05] + sf_payload)	
		self.simulate_rx(data = [0x10, payload2_size] + payload2[0:6])
		self.stack.process()
		self.simulate_rx(data = [0x21] + payload2[6:13])		
		self.simulate_rx(data = [0x22] + payload2[13:16])		
		self.stack.process()
		self.assertEqual(self.rx_isotp_frame(), bytearray(sf_payload))
		self.assertEqual(self.rx_isotp_frame(), bytearray(payload2))
		self.assertIsNone(self.rx_isotp_frame())
		self.assert_error_triggered(isotp.ReceptionInterruptedWithSingleFrameError)

	def test_receive_4095_multiframe(self):
		payload_size = 4095
		payload = self.make_payload(payload_size)
		self.simulate_rx(data = [0x1F, 0xFF] + payload[0:6])
		n = 6
		seqnum = 1
		while n<4096:
			self.simulate_rx(data = [0x20 | (seqnum & 0xF)] + payload[n:min(n+7, 4096)])	
			self.stack.process()
			n += 7
			seqnum +=1
		self.assertEqual(self.rx_isotp_frame(), bytearray(payload))
		self.assertIsNone(self.rx_isotp_frame())

	def test_receive_4095_multiframe_check_blocksize(self):
		for blocksize in range(1,10):
			self.perform_receive_4095_multiframe_check_blocksize(blocksize=blocksize)

	def perform_receive_4095_multiframe_check_blocksize(self, blocksize):
		payload_size = 4095
		self.stack.params.set('blocksize', blocksize)
		self.stack.params.set('stmin', 2)

		payload = self.make_payload(payload_size)
		self.simulate_rx(data = [0x1F, 0xFF] + payload[0:6])
		self.stack.process()
		self.assert_sent_flow_control(blocksize=blocksize, stmin=2, extra_msg='blocksize=%d' % blocksize)
		n = 6
		block_counter = 0
		seqnum = 1
		while n<4096:
			self.simulate_rx(data = [0x20 | (seqnum & 0xF)] + payload[n:min(n+7, 4096)])	
			self.stack.process()
			block_counter+=1
			n += 7
			seqnum +=1
			if block_counter % blocksize == 0 and n < 4095:
				self.assert_sent_flow_control(blocksize=blocksize, stmin=2, extra_msg='blocksize=%d' % blocksize)
			else:
				self.assertIsNone(self.get_tx_can_msg(), 'Sent a message something after block %d but shoud not have. blocksize = %d' % (block_counter, blocksize))

		self.assertEqual(self.rx_isotp_frame(), bytearray(payload), 'blocksize=%d' % blocksize)
		self.assertIsNone(self.rx_isotp_frame(), 'blocksize=%d' % blocksize)

	def receive_invalid_can_message(self):
		for i in range(4, 0x10):
			self.simulate_rx(data = [i << 4, 0x00, 0x00])
			self.stack.process()
			self.assert_error_triggered(isotp.InvalidCanDataError)
			self.clear_errors()

	def test_receive_data_length_12_bytes(self):
		self.stack.params.set('ll_data_length', 12)
		payload_size = 30
		payload = self.make_payload(payload_size)
		self.simulate_rx(data = [0x10, payload_size] + payload[0:10])
		self.stack.process()
		self.simulate_rx(data = [0x21] + payload[10:21])
		self.simulate_rx(data = [0x22] + payload[21:30])
		self.stack.process()
		data = self.rx_isotp_frame()
		self.assertEqual(data, bytearray(payload))

	def test_receive_data_length_5_bytes(self):
		self.stack.params.set('ll_data_length', 5)
		payload_size = 15
		payload = self.make_payload(payload_size)
		self.simulate_rx(data = [0x10, payload_size] + payload[0:3])
		self.stack.process()
		self.simulate_rx(data = [0x21] + payload[3:7])
		self.simulate_rx(data = [0x22] + payload[7:11])
		self.simulate_rx(data = [0x23] + payload[11:15])
		self.stack.process()
		data = self.rx_isotp_frame()
		self.assertEqual(data, bytearray(payload))

	def test_receive_data_length_12_but_set_8_bytes(self):
		self.stack.params.set('ll_data_length', 8)
		payload_size = 30
		payload = self.make_payload(payload_size)
		self.simulate_rx(data = [0x10, payload_size] + payload[0:10])
		self.stack.process()
		self.simulate_rx(data = [0x21] + payload[10:21])
		self.simulate_rx(data = [0x22] + payload[21:30])
		self.stack.process()
		data = self.rx_isotp_frame()
		self.assertNotEqual(data, bytearray(payload))


	# ================ Transmission ====================

	def assert_tx_timing_spin_wait_for_msg(self, mintime, maxtime):
		msg = None
		diff = 0
		t = time.time()
		while msg is None:
			self.stack.process()
			msg = self.get_tx_can_msg()
			diff = time.time() - t
			self.assertLess(diff, maxtime, 'Timed out') # timeout
		self.assertGreater(diff, mintime, 'Stack sent a message too quickly')
		return msg


	def test_send_single_frame(self):
		for i in range(1,7):
			payload = self.make_payload(i,i)
			self.assertIsNone(self.get_tx_can_msg())
			self.tx_isotp_frame(payload)
			self.stack.process()
			msg = self.get_tx_can_msg()
			self.assertEqual(msg.arbitration_id, self.TXID)
			self.assertEqual(msg.dlc, i+1)
			self.assertEqual(msg.data, bytearray([0x0 | i] + payload))

	def test_padding_single_frame(self):
		padding_byte = 0xAA
		self.stack.params.set('tx_padding', padding_byte)

		for i in range(1,7):
			payload = self.make_payload(i,i)
			self.assertIsNone(self.get_tx_can_msg())
			self.tx_isotp_frame(payload)
			self.stack.process()
			msg = self.get_tx_can_msg()
			self.assertEqual(msg.arbitration_id, self.TXID)
			self.assertEqual(msg.dlc, 8)
			self.assertEqual(msg.data, bytearray([0x0 | i] + payload + [padding_byte] * (7-i)))

	def test_padding_single_frame_dl_12_bytes(self):
		padding_byte = 0xAA
		self.stack.params.set('tx_padding', padding_byte)
		self.stack.params.set('ll_data_length', 12)

		for i in range(1,11):
			payload = self.make_payload(i,i)
			self.assertIsNone(self.get_tx_can_msg())
			self.tx_isotp_frame(payload)
			self.stack.process()
			msg = self.get_tx_can_msg()
			self.assertEqual(msg.arbitration_id, self.TXID)
			self.assertEqual(msg.dlc, 12)
			self.assertEqual(msg.data, bytearray([0x0 | i] + payload + [padding_byte] * (11-i)))

	def test_send_multiple_single_frame_one_process(self):
		payloads = dict()
		for i in range(1,8):
			payload = self.make_payload(i,i)
			self.tx_isotp_frame(payload)
			payloads[i] = payload

		self.stack.process()

		for i in range(1,8):
			msg = self.get_tx_can_msg()
			self.assertIsNotNone(msg)
			self.assertEqual(msg.arbitration_id, self.TXID)
			self.assertEqual(msg.dlc, i+1)
			self.assertEqual(msg.data, bytearray([0x0 | i] + payloads[i]))

	def test_send_small_multiframe(self):
		payload = self.make_payload(10)
		self.assertIsNone(self.get_tx_can_msg())
		self.tx_isotp_frame(payload)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertEqual(msg.arbitration_id, self.TXID)
		self.assertEqual(msg.dlc, 8)
		self.assertEqual(msg.data, bytearray([0x10, 0x0A] + payload[:6]))
		self.assertIsNone(self.get_tx_can_msg())
		self.stack.process()
		self.assertIsNone(self.get_tx_can_msg())
		self.simulate_rx_flowcontrol(flow_status=0, stmin=0, blocksize=8)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertIsNotNone(msg)
		self.assertEqual(msg.arbitration_id, self.TXID)
		self.assertEqual(msg.dlc, 5)
		self.assertEqual(msg.data, bytearray([0x21] + payload[6:10]))
		self.assertIsNone(self.get_tx_can_msg())
		self.stack.process()
		self.assertIsNone(self.get_tx_can_msg())

	def test_padding_multi_frame(self):
		padding_byte = 0x55
		self.stack.params.set('tx_padding', padding_byte)
		payload = self.make_payload(10)
		self.assertIsNone(self.get_tx_can_msg())
		self.tx_isotp_frame(payload)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertEqual(msg.arbitration_id, self.TXID)
		self.assertEqual(msg.dlc, 8)
		self.assertEqual(msg.data, bytearray([0x10, 0x0A] + payload[:6]))
		self.assertIsNone(self.get_tx_can_msg())
		self.stack.process()
		self.assertIsNone(self.get_tx_can_msg())
		self.simulate_rx_flowcontrol(flow_status=0, stmin=0, blocksize=8)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertIsNotNone(msg)
		self.assertEqual(msg.arbitration_id, self.TXID)
		self.assertEqual(msg.dlc, 8)
		self.assertEqual(msg.data, bytearray([0x21] + payload[6:10] + [padding_byte]*3))
		self.assertIsNone(self.get_tx_can_msg())
		self.stack.process()
		self.assertIsNone(self.get_tx_can_msg())

	def test_padding_multi_frame_dl_12_bytes(self):
		padding_byte = 0x55
		self.stack.params.set('tx_padding', padding_byte)
		self.stack.params.set('ll_data_length', 12)
		payload = self.make_payload(15)
		self.assertIsNone(self.get_tx_can_msg())
		self.tx_isotp_frame(payload)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertEqual(msg.arbitration_id, self.TXID)
		self.assertEqual(msg.dlc, 12)
		self.assertEqual(msg.data, bytearray([0x10, 15] + payload[:10]))
		self.assertIsNone(self.get_tx_can_msg())
		self.stack.process()
		self.assertIsNone(self.get_tx_can_msg())
		self.simulate_rx_flowcontrol(flow_status=0, stmin=0, blocksize=8)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertIsNotNone(msg)
		self.assertEqual(msg.arbitration_id, self.TXID)
		self.assertEqual(msg.dlc, 12)
		self.assertEqual(msg.data, bytearray([0x21] + payload[10:15] + [padding_byte]*6))
		self.assertIsNone(self.get_tx_can_msg())
		self.stack.process()
		self.assertIsNone(self.get_tx_can_msg())

	def test_send_2_small_multiframe(self):
		payload1 = self.make_payload(10)
		payload2 = self.make_payload(10,1)
		self.tx_isotp_frame(payload1)
		self.tx_isotp_frame(payload2)

		self.assertIsNone(self.get_tx_can_msg())
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertIsNotNone(msg)
		self.assertEqual(msg.data, bytearray([0x10, 0x0A] + payload1[:6]))
		self.assertIsNone(self.get_tx_can_msg())
		self.stack.process()
		self.assertIsNone(self.get_tx_can_msg())
		self.simulate_rx_flowcontrol(flow_status=0, stmin=0, blocksize=8)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertIsNotNone(msg)
		self.assertEqual(msg.data, bytearray([0x21] + payload1[6:10]))
		msg = self.get_tx_can_msg()
		self.assertIsNotNone(msg)
		self.assertEqual(msg.data, bytearray([0x10, 0x0A] + payload2[:6]))
		self.assertIsNone(self.get_tx_can_msg())
		self.stack.process()
		self.assertIsNone(self.get_tx_can_msg())
		self.simulate_rx_flowcontrol(flow_status=0, stmin=0, blocksize=8)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertIsNotNone(msg)
		self.assertEqual(msg.data, bytearray([0x21] + payload2[6:10]))
		self.assertIsNone(self.get_tx_can_msg())

	def test_send_multiframe_flow_control_timeout(self):
		self.stack.params.set('rx_flowcontrol_timeout', 200)

		payload = self.make_payload(10)
		self.tx_isotp_frame(payload)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertEqual(msg.data, bytearray([0x10, 0x0A] + payload[:6]))
		self.assertIsNone(self.get_tx_can_msg())
		self.stack.process()
		self.assertIsNone(self.get_tx_can_msg())
		time.sleep(0.2)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertIsNone(msg)
		self.assert_error_triggered(isotp.FlowControlTimeoutError)

	def test_send_multiframe_flow_control_timeout_recover(self):
		self.stack.params.set('rx_flowcontrol_timeout', 200)

		payload = self.make_payload(10)
		payload2 = self.make_payload(10,1)
		self.tx_isotp_frame(payload)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertEqual(msg.data, bytearray([0x10, 0x0A] + payload[:6]))
		self.assertIsNone(self.get_tx_can_msg())
		self.stack.process()
		self.assertIsNone(self.get_tx_can_msg())
		time.sleep(0.2)
		self.tx_isotp_frame(payload)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertEqual(msg.data, bytearray([0x10, 0x0A] + payload[:6]))
		self.simulate_rx_flowcontrol(flow_status=0, stmin=0, blocksize=8)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertEqual(msg.data, bytearray([0x21] + payload[6:10]))
		self.assert_error_triggered(isotp.FlowControlTimeoutError)

	def test_send_unexpected_flow_control(self):
		self.simulate_rx_flowcontrol(flow_status=0, stmin=100, blocksize=8)
		self.stack.process()
		self.assertIsNone(self.get_tx_can_msg())
		self.assert_error_triggered(isotp.UnexpectedFlowControlError)

	def test_send_respect_wait_frame(self):
		self.stack.params.set('wftmax', 15)
		self.stack.params.set('rx_flowcontrol_timeout', 500)

		payload = self.make_payload(20)
		self.tx_isotp_frame(payload)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertEqual(msg.data, bytearray([0x10, 20] + payload[:6]))
		for i in range(10):
			self.simulate_rx_flowcontrol(flow_status=1, stmin=0, blocksize=1)
			self.stack.process()
			time.sleep(0.2)
		self.simulate_rx_flowcontrol(flow_status=0, stmin=0, blocksize=1)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertEqual(msg.data, bytearray([0x21] + payload[6:13]))
		for i in range(10):
			self.simulate_rx_flowcontrol(flow_status=1, stmin=0, blocksize=1)
			self.stack.process()
			time.sleep(0.2)
		self.simulate_rx_flowcontrol(flow_status=0, stmin=0, blocksize=1)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertEqual(msg.data, bytearray([0x22] + payload[13:20]))

		self.assert_no_error_triggered()

	def test_send_respect_wait_frame_but_timeout(self):
		self.stack.params.set('wftmax', 15)
		self.stack.params.set('rx_flowcontrol_timeout', 500)

		payload = self.make_payload(20)
		self.tx_isotp_frame(payload)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertEqual(msg.data, bytearray([0x10, 20] + payload[:6]))
		for i in range(3):
			self.simulate_rx_flowcontrol(flow_status=1, stmin=0, blocksize=1)
			self.stack.process()
			time.sleep(0.2)
		time.sleep(0.5)
		self.simulate_rx_flowcontrol(flow_status=0, stmin=0, blocksize=1)
		self.stack.process()
		self.assertIsNone(self.get_tx_can_msg())
		
		self.assert_error_triggered(isotp.FlowControlTimeoutError)

	def test_send_wait_frame_after_first_frame_wftmax_0(self):
		self.stack.params.set('wftmax', 0)
		payload = self.make_payload(10)
		self.tx_isotp_frame(payload)
		self.stack.process()
		self.simulate_rx_flowcontrol(flow_status=1, stmin=0, blocksize=8)
		self.stack.process()
		time.sleep(0.01)
		self.simulate_rx_flowcontrol(flow_status=0, stmin=0, blocksize=8)
		self.stack.process()
		self.assert_error_triggered(isotp.UnsuportedWaitFrameError)

	def test_send_wait_frame_after_consecutive_frame_wftmax_0(self):
		self.stack.params.set('wftmax', 0)

		payload = self.make_payload(20)
		self.tx_isotp_frame(payload)
		self.stack.process()
		self.simulate_rx_flowcontrol(flow_status=0, stmin=0, blocksize=1)
		self.stack.process()
		self.simulate_rx_flowcontrol(flow_status=1, stmin=0, blocksize=1)
		self.stack.process()
		self.simulate_rx_flowcontrol(flow_status=0, stmin=0, blocksize=1)
		self.stack.process()
		self.assert_error_triggered(isotp.UnsuportedWaitFrameError)

	def test_send_wait_frame_after_first_frame_reach_max(self):
		self.stack.params.set('wftmax', 5)

		payload = self.make_payload(20)
		self.tx_isotp_frame(payload)
		self.stack.process()
		self.get_tx_can_msg()
		for i in range(6):
			self.simulate_rx_flowcontrol(flow_status=1, stmin=0, blocksize=1)
			self.stack.process()
			time.sleep(0.2)
		self.simulate_rx_flowcontrol(flow_status=0, stmin=0, blocksize=1)
		self.stack.process()
		self.assertIsNone(self.get_tx_can_msg())
		self.assert_error_triggered(isotp.MaximumWaitFrameReachedError)
		self.assert_error_triggered(isotp.UnexpectedFlowControlError)

	def test_send_wait_frame_after_conscutive_frame_reach_max(self):
		self.stack.params.set('wftmax', 5)

		payload = self.make_payload(20)
		self.tx_isotp_frame(payload)
		self.stack.process()
		self.get_tx_can_msg()
		self.simulate_rx_flowcontrol(flow_status=0, stmin=0, blocksize=1)
		self.stack.process()
		self.get_tx_can_msg()
		for i in range(6):
			self.simulate_rx_flowcontrol(flow_status=1, stmin=0, blocksize=1)
			self.stack.process()
			time.sleep(0.2)
		self.simulate_rx_flowcontrol(flow_status=0, stmin=0, blocksize=1)
		self.stack.process()
		self.assertIsNone(self.get_tx_can_msg())
		self.assert_error_triggered(isotp.MaximumWaitFrameReachedError)
		self.assert_error_triggered(isotp.UnexpectedFlowControlError)

	def test_send_4095_multiframe_zero_stmin(self):
		self.perform_multiframe_test_no_stmin(4095, 5)

	def test_send_128_multiframe_variable_blocksize(self):
		for i in range(1,8):
			self.perform_multiframe_test_no_stmin(128, i)
	
	def perform_multiframe_test_no_stmin(self, payload_size, blocksize):
		stmin = 0
		payload = self.make_payload(payload_size)
		self.tx_isotp_frame(payload)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertIsNotNone(msg, 'blocksize = %d' % blocksize)
		self.assertEqual(msg.data, bytearray([0x10 | ((payload_size >> 8) & 0xF), payload_size & 0xFF] + payload[:6]), 'blocksize = %d' % blocksize)
		self.simulate_rx_flowcontrol(flow_status=0, stmin=stmin, blocksize=blocksize)

		block_counter = 0
		seqnum = 1
		n = 6
		self.stack.process()	# Call only once, should enqueue all message until next flow control
		while True:
			msg = self.get_tx_can_msg()
			if block_counter < blocksize:
				self.assertIsNotNone(msg, 'blocksize = %d' % blocksize)
				self.assertEqual(msg.data, bytearray([0x20 | seqnum] + payload[n:min(n+7, payload_size)]), 'blocksize = %d' % blocksize)
				n+=7
				seqnum = (seqnum+1) & 0xF
				block_counter+=1
				if n > payload_size:
					break
			else:
				self.assertIsNone(msg, 'blocksize = %d' % blocksize)
				self.simulate_rx_flowcontrol(flow_status=0, stmin=stmin, blocksize=blocksize)
				self.stack.process() # Receive the flow control and enqueue another bloc of can message. 
				block_counter = 0

	def test_squash_timing_requirement(self):
		self.stack.params.set('squash_stmin_requirement', True)

		payload_size = 4095
		stmin=100 # 100 msec
		blocksize = 8

		payload = self.make_payload(payload_size)
		self.tx_isotp_frame(payload)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertIsNotNone(msg)
		self.assertEqual(msg.data, bytearray([0x10 | ((payload_size >> 8) & 0xF), payload_size & 0xFF] + payload[:6]))
		self.simulate_rx_flowcontrol(flow_status=0, stmin=stmin, blocksize=blocksize)

		block_counter = 0
		seqnum = 1
		n = 6
		self.stack.process()	# Call only once, should enqueue all message until next flow control
		while True:
			msg = self.get_tx_can_msg()
			if block_counter < blocksize:
				self.assertIsNotNone(msg)
				self.assertEqual(msg.data, bytearray([0x20 | seqnum] + payload[n:min(n+7, payload_size)]))
				n+=7
				seqnum = (seqnum+1) & 0xF
				block_counter+=1
				if n > payload_size:
					break
			else:
				self.assertIsNone(msg)
				self.simulate_rx_flowcontrol(flow_status=0, stmin=stmin, blocksize=blocksize)
				self.stack.process() # Receive the flow control and enqueue another bloc of can message. 
				block_counter = 0

	def test_stmin_requirement(self):
		stmin = 100 # 100 msec
		payload_size = 30
		blocksize = 3
		payload = self.make_payload(payload_size)
		self.tx_isotp_frame(payload)
		self.stack.process()
		t = time.time()
		msg = self.get_tx_can_msg()
		self.assertEqual(msg.data, bytearray([0x10 | ((payload_size >> 8) & 0xF), payload_size & 0xFF] + payload[:6]), 'stmin = %d' % stmin)
		self.simulate_rx_flowcontrol(flow_status=0, stmin=stmin, blocksize=blocksize)
		msg = self.assert_tx_timing_spin_wait_for_msg(mintime=0.095, maxtime=1)
		self.assertEqual(msg.data, bytearray([0x21] + payload[6:13]))
		msg = self.assert_tx_timing_spin_wait_for_msg(mintime=0.095, maxtime=1)
		self.assertEqual(msg.data, bytearray([0x22] + payload[13:20]))
		msg = self.assert_tx_timing_spin_wait_for_msg(mintime=0.095, maxtime=1)
		self.assertEqual(msg.data, bytearray([0x23] + payload[20:27]))
		self.simulate_rx_flowcontrol(flow_status=0, stmin=stmin, blocksize=blocksize)
		msg = self.assert_tx_timing_spin_wait_for_msg(mintime=0.095, maxtime=1)
		self.assertEqual(msg.data, bytearray([0x24] + payload[27:30]))

	def test_send_nothing_with_empty_payload(self):
		self.tx_isotp_frame([])
		self.tx_isotp_frame([])
		self.tx_isotp_frame([])
		self.stack.process()
		self.assertIsNone(self.get_tx_can_msg())

	def test_send_single_frame_after_empty_payload(self):
		self.tx_isotp_frame([])
		self.tx_isotp_frame([0x55])
		self.tx_isotp_frame([])
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertIsNotNone(msg)
		self.assertEqual(msg.data, bytearray([0x01, 0x55]))

	#Sets blocksize to 0, never sends flow control except after first frame
	def test_send_blocksize_zero(self):
		payload = self.make_payload(4095)
		self.tx_isotp_frame(payload)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertEqual(msg.data, bytearray([0x1F, 0xFF] + payload[:6]))
		self.simulate_rx_flowcontrol(flow_status=0, stmin=0, blocksize=0)

		seqnum = 1
		n=6
		self.stack.process()
		while True:
			msg = self.get_tx_can_msg()
			self.assertIsNotNone(msg)
			self.assertEqual(msg.data, bytearray([0x20 | seqnum] + payload[n:min(n+7, 4095)]))
			n+=7
			seqnum = (seqnum+1) & 0xF

			if n > 4095:
				break

	def test_transmit_data_length_12_bytes(self):
		self.stack.params.set('ll_data_length', 12)
		payload = self.make_payload(30)
		self.tx_isotp_frame(payload)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertEqual(msg.data, bytearray([0x10, 30] + payload[:10]))
		self.simulate_rx_flowcontrol(flow_status=0, stmin=0, blocksize=0)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertEqual(msg.data, bytearray([0x21] + payload[10:21]))
		msg = self.get_tx_can_msg()
		self.assertEqual(msg.data, bytearray([0x22] + payload[21:30]))

	def test_transmit_data_length_5_bytes(self):
		self.stack.params.set('ll_data_length', 5)
		payload = self.make_payload(15)
		self.tx_isotp_frame(payload)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertEqual(msg.data, bytearray([0x10, 15] + payload[:3]))
		self.simulate_rx_flowcontrol(flow_status=0, stmin=0, blocksize=0)
		self.stack.process()
		msg = self.get_tx_can_msg()
		self.assertEqual(msg.data, bytearray([0x21] + payload[3:7]))
		msg = self.get_tx_can_msg()
		self.assertEqual(msg.data, bytearray([0x22] + payload[7:11]))
		msg = self.get_tx_can_msg()
		self.assertEqual(msg.data, bytearray([0x23] + payload[11:15]))


	# =============== Parameters ===========
	def create_layer(self, params):
		address = isotp.Address(isotp.AddressingMode.Normal_11bits, txid=self.TXID, rxid=self.RXID)
		return isotp.TransportLayer(txfn=self.stack_txfn, rxfn=self.stack_rxfn, address=address, params = params)

	def test_params_bad_values(self):
		params = {
			'stmin' : 1,
			'blocksize' : 8,
			'squash_stmin_requirement' : False,
			'rx_flowcontrol_timeout'  : 1000,
			'rx_consecutive_frame_timeout' : 1000
		}

		self.create_layer({}) # Empty params. Use default value
		self.create_layer(params)

		with self.assertRaises(ValueError):
			params['stmin'] = -1
			self.create_layer(params)

		with self.assertRaises(ValueError):
			params['stmin'] = 0x100
			self.create_layer(params)

		with self.assertRaises(ValueError):
			params['stmin'] = 'string'
			self.create_layer(params)
		params['stmin'] = 1


		with self.assertRaises(ValueError):
			params['blocksize'] = -1
			self.create_layer(params)

		with self.assertRaises(ValueError):
			params['blocksize'] = 0x100
			self.create_layer(params)

		with self.assertRaises(ValueError):
			params['blocksize'] = 'string'
			self.create_layer(params)
		params['blocksize'] = 8


		with self.assertRaises(ValueError):
			params['squash_stmin_requirement'] = 'string'
			self.create_layer(params)
		params['squash_stmin_requirement'] = False


		with self.assertRaises(ValueError):
			params['rx_flowcontrol_timeout'] = -1
			self.create_layer(params)

		with self.assertRaises(ValueError):
			params['rx_flowcontrol_timeout'] = 'string'
			self.create_layer(params)
		params['rx_flowcontrol_timeout'] = 1000


		with self.assertRaises(ValueError):
			params['rx_consecutive_frame_timeout'] = -1
			self.create_layer(params)

		with self.assertRaises(ValueError):
			params['rx_consecutive_frame_timeout'] = 'string'
			self.create_layer(params)
		params['rx_consecutive_frame_timeout'] = 1000





import socket as socket_module
import os
import isotp.address

mtu=4095

def check_support():
    if not hasattr(socket_module, 'CAN_ISOTP'):
        if os.name == 'nt':
            raise NotImplementedError("This module cannot be used on Windows")
        else:
            raise NotImplementedError("Your version of Python does not offer support for CAN ISO-TP protocol. Support have been added since Python 3.7 on Linux build > 2.6.15.")

class flags:
        
    LISTEN_MODE     = 0x001
    """Puts the socket in Listen mode, which prevents transmission of data"""

    EXTEND_ADDR     = 0x002
    """When set, an address extension byte (set in socket general options) will be added to each payload sent. Unless RX_EXT_ADDR is also set, this value will be expected for reception as well"""

    TX_PADDING      = 0x004
    """Enables padding of transmitted data with a byte set in the socket general options"""

    RX_PADDING      = 0x008
    """ Indicates that data padding is possible in reception. Must be set for CHK_PAD_LEN and CHK_PAD_DATA to have an effect"""

    CHK_PAD_LEN     = 0x010
    """ Makes the socket validate the padding length of the CAN message"""

    CHK_PAD_DATA    = 0x020
    """ Makes the socket validate the padding bytes of the CAN message"""

    HALF_DUPLEX     = 0x040
    """ Sets the socket in half duplex mode, forcing transmission and reception to happen sequentially """

    FORCE_TXSTMIN   = 0x080
    """Forces the socket to use the separation time sets in general options, overriding stmin value received in flow control frames."""
    
    FORCE_RXSTMIN   = 0x100
    """ Forces the socket to ignore any message received faster than stmin given in the flow control frame"""

    RX_EXT_ADDR     = 0x200
    """ When sets, a different extended address can be used for reception than for transmission."""

class LinkLayerProtocol:
    CAN = 16
    """ Internal structure size of a CAN 2.0 frame"""

    CAN_FD = 72
    """ Internal structure size of a CAN FD frame"""

class socket:
    """
    A IsoTP socket wrapper for easy configuration

    :param timeout: The underlying socket timeout set with ``settimeout``. Makes the reception thread sleep
    :type timeout: int

    """

    # We want that syntax isotp.socket.flags and isotp.socket.mtu
    # This is a workaround for sphinx autodoc that fails to load docstring for nested-class members
    flags = flags   
    LinkLayerProtocol = LinkLayerProtocol

    def __init__(self, timeout=0.1):
        check_support()
        from . import opts
        self.interface = None
        self.address = None
        self.bound = False
        self.closed = False
        self._socket = socket_module.socket(socket_module.AF_CAN, socket_module.SOCK_DGRAM, socket_module.CAN_ISOTP)
        if timeout is not None and timeout>0:
            self._socket.settimeout(timeout)

    def send(self, *args, **kwargs):
        if not self.bound:
            raise RuntimeError("bind() must be called before using the socket")
        return self._socket.send(*args, **kwargs)

    def recv(self, n=mtu):
        if not self.bound:
            raise RuntimeError("bind() must be called before using the socket")
        try:
            return self._socket.recv(n)
        except socket_module.timeout:
            return None
        except:
            raise

    def set_ll_opts(self, *args, **kwargs):
        if self.bound:
            raise RuntimeError("Options must be set before calling bind()")
        return opts.linklayer.write(self._socket, *args, **kwargs)

    def set_opts(self, *args, **kwargs):
        if self.bound:
            raise RuntimeError("Options must be set before calling bind()")
        return opts.general.write(self._socket, *args, **kwargs)

    def set_fc_opts(self, *args, **kwargs):
        if self.bound:
            raise RuntimeError("Options must be set before calling bind()")
        return opts.flowcontrol.write(self._socket, *args, **kwargs)

    def get_ll_opts(self, *args, **kwargs):
        return opts.linklayer.read(self._socket, *args, **kwargs)

    def get_opts(self, *args, **kwargs):
        return opts.general.read(self._socket, *args, **kwargs)

    def get_fc_opts(self, *args, **kwargs):
        return opts.flowcontrol.read(self._socket, *args, **kwargs)

    def bind(self, interface, *args, **kwargs):
        """
        Binds the socket to an address. 
        If no address is provided, all additional parameters will be used to create an adresse. This is mainly to allow a syntax such as ``sock.bind('vcan0', rxid=0x123, txid=0x456)`` for backward compatibility.

        :param interface: The network interface to use
        :type interface: string

        :param address: The address to bind to. 
        :type: :class:`isotp.Address<isotp.Address>`
        """
        self.interface=interface

        # == This is for syntax flexibility and also backward compatibility
        address = None
        if 'address' in kwargs:
            address = kwargs['address']

        for arg in args:
            if isinstance(arg, isotp.address.Address) and address is None:
                address = arg
                break

        if address is None:
            address = isotp.address.Address(*args, **kwargs)
        # == 
        self.address = address

        # IsoTP sockets doesn't provide an interface to modify the target address type. We asusme physical.
        # If functional is required, it Ids can be manually crafted in Normal / extended mode
        rxid = self.address.get_rx_arbitraton_id(isotp.TargetAddressType.Physical)
        txid = self.address.get_tx_arbitraton_id(isotp.TargetAddressType.Physical)
        
        if self.address.is_29bits == True:
            rxid = (rxid & socket_module.CAN_EFF_MASK) | socket_module.CAN_EFF_FLAG
        else:
            rxid = rxid & socket_module.CAN_SFF_MASK

        if self.address.is_29bits == True:
            txid = (txid & socket_module.CAN_EFF_MASK) | socket_module.CAN_EFF_FLAG
        else:
            txid = txid & socket_module.CAN_SFF_MASK

        if self.address.requires_extension_byte():
            o = self.get_opts()
            o.optflag |= self.flags.EXTEND_ADDR | self.flags.RX_EXT_ADDR
            self.set_opts(optflag = o.optflag, ext_address = self.address.get_tx_extension_byte(), rx_ext_address=self.address.get_rx_extension_byte())
            
        self._socket.bind((interface, rxid, txid))
        self.bound=True

    def fileno(self):
        return self._socket.fileno()

    def close(self, *args, **kwargs):
        v = self._socket.close(*args, **kwargs)
        self.bound = False
        self.closed = True
        self.address = None
        return v

    def __delete__(self):
        if isinstance(_socket, socket_module.socket):
            self._socket.close()
            self._socket = None

    def __repr__(self):
        if self.bound:
            return "<ISO-TP Socket: %s, %s>" % (self.interface, self.address.get_content_str())
        else:
            status = "Closed" if self.closed else "Unbound"
            return "<%s ISO-TP Socket at 0x%s>" % (status, hex(id(self)))



import unittest 
import queue
import _thread as thread
import threading

#Class borrowed from Python Socket test suite.
class ThreadableTest(unittest.TestCase):
	def __init__(self, *args, **kwargs):
		unittest.TestCase.__init__(self, *args, **kwargs)
		# Swap the true setup function
		self.__setUp = self.setUp
		self.__tearDown = self.tearDown
		self.setUp = self._setUp
		self.tearDown = self._tearDown

	def serverExplicitReady(self):
		self.server_ready.set()

	def _setUp(self):
		self.server_ready = threading.Event()
		self.client_ready = threading.Event()
		self.done = threading.Event()
		self.queue = queue.Queue(1)
		self.server_crashed = False

		# Do some munging to start the client test.
		methodname = self.id()
		i = methodname.rfind('.')
		methodname = methodname[i+1:]
		test_method = getattr(self, '_' + methodname)
		self.client_thread = thread.start_new_thread(self.clientRun, (test_method,))

		try:
			self.__setUp()
		except:
			self.server_crashed = True
			raise
		finally:
			self.server_ready.set()
		self.client_ready.wait()

	def _tearDown(self):
		self.__tearDown()
		self.done.wait()

		if self.queue.qsize():
			exc = self.queue.get()
			raise exc

	def clientRun(self, test_func):
		self.server_ready.wait()
		try:
			self.clientSetUp()
		except BaseException as e:
			self.queue.put(e)
			self.client_ready.set()
			self._clientTearDown()
			return
		finally:
			self.client_ready.set()

		if self.server_crashed:
			self._clientTearDown()
			return
		if not hasattr(test_func, '__call__'):
			raise TypeError("test_func must be a callable function")
		try:
			test_func()
		except BaseException as e:
			self.queue.put(e)
		finally:
			self._clientTearDown()

	def clientSetUp(self):
		raise NotImplementedError("clientSetUp must be implemented.")

	def _clientTearDown(self):
		self.done.set()
		try:
			if hasattr(self, 'clientTearDown'):
				self.clientTearDown()
		except BaseException as e:
			self.queue.put(e)
		finally:			
			thread.exit()


import queue
import logging
from copy import copy
import binascii
import time
import functools
import isotp.address
import isotp.errors

class CanMessage:
	"""
	Represent a CAN message (ISO-11898)

	:param arbitration_id: The CAN arbitration ID. Must be a 11 bits value or a 29 bits value if ``extended_id`` is True
	:type arbitration_id: int

	:param dlc: The Data Length Code representing the number of bytes in the data field
	:type dlc: int

	:param data: The 8 bytes payload of the message
	:type data: bytearray

	:param extended_id: When True, the arbitration ID stands on 29 bits. 11 bits when False
	:type extended_id: bool
	"""
	__slots__ = 'arbitration_id', 'dlc', 'data','is_extended_id'

	def __init__(self, arbitration_id=None, dlc=None, data=None, extended_id=False):
		self.arbitration_id = arbitration_id
		self.dlc = dlc
		self.data = data
		self.is_extended_id = extended_id

class PDU:
	"""
	Converts a CAN Message into a meaningful PDU such as SingleFrame, FirstFrame, ConsecutiveFrame, FlowControl

	:param msg: The CAN message
	:type msg: `isotp.protocol.CanMessage`
	"""
	__slots__ = 'type', 'length', 'data', 'blocksize', 'stmin', 'stmin_sec', 'seqnum', 'flow_status'

	class Type:
		SINGLE_FRAME = 0
		FIRST_FRAME = 1
		CONSECUTIVE_FRAME = 2
		FLOW_CONTROL = 3

	class FlowStatus:
		ContinueToSend = 0
		Wait = 1
		Overflow = 2

	def __init__(self, msg = None, start_of_data=0, datalen=8):
		
		self.type = None
		self.length = None
		self.data = None
		self.blocksize = None
		self.stmin = None
		self.stmin_sec = None
		self.seqnum = None
		self.flow_status = None

		if msg is None:
			return

		if len(msg.data)>start_of_data:
			hnb =  (msg.data[start_of_data] >> 4) & 0xF
			if hnb > 3:
				raise ValueError('Received message with unknown frame type %d' % hnb)
			self.type = int(hnb)
		else:
			raise ValueError('Empty CAN frame')

		if self.type == self.Type.SINGLE_FRAME:
			self.length = int(msg.data[start_of_data]) & 0xF
			if len(msg.data) < self.length + 1:
				raise ValueError('Single Frame length is bigger than CAN frame length')

			if self.length == 0 or self.length > datalen-1-start_of_data:
				raise ValueError("Received Single Frame with invalid length of %d" % self.length)
			self.data = msg.data[1+start_of_data:][:self.length]

		elif self.type == self.Type.FIRST_FRAME:
			if len(msg.data) < 2+start_of_data:
				raise ValueError('First frame must be at least %d bytes long' % (2+start_of_data))

			self.length = ((int(msg.data[start_of_data]) & 0xF) << 8) | int(msg.data[start_of_data+1])
			if len(msg.data) < datalen:
				if len(msg.data) < self.length + 2 + start_of_data:
					raise ValueError('First frame specifies a length that is inconsistent with underlying CAN message DLC')

			self.data = msg.data[2+start_of_data:][:min(self.length, datalen-2-start_of_data)]
				
		elif self.type == self.Type.CONSECUTIVE_FRAME:
			self.seqnum = int(msg.data[start_of_data]) & 0xF
			self.data = msg.data[start_of_data+1:datalen]

		elif self.type == self.Type.FLOW_CONTROL:
			if len(msg.data) < 3+start_of_data:
				raise ValueError('Flow Control frame must be at least 3 bytes')

			self.flow_status = int(msg.data[start_of_data]) & 0xF
			if self.flow_status >= 3:
				raise ValueError('Unknown flow status')

			self.blocksize = int(msg.data[1+start_of_data])
			stmin_temp = int(msg.data[2+start_of_data])

			if stmin_temp >= 0 and stmin_temp <= 0x7F:
				self.stmin_sec = stmin_temp / 1000
			elif stmin_temp >= 0xf1 and stmin_temp <= 0xF9:
				self.stmin_sec = (stmin_temp - 0xF0) / 10000

			if self.stmin_sec is None:
				raise ValueError('Invalid StMin received in Flow Control')
			else:
				self.stmin = stmin_temp
	@classmethod
	def craft_flow_control_data(cls, flow_status, blocksize, stmin):
		return bytearray([ (0x30 | (flow_status) & 0xF), blocksize&0xFF, stmin & 0xFF])

	def name(self):
		if self.type is None:
			return "[None]"

		if self.type == self.Type.SINGLE_FRAME:
			return "SINGLE_FRAME"
		elif self.type == self.Type.FIRST_FRAME:
			return "FIRST_FRAME"
		elif self.type == self.Type.CONSECUTIVE_FRAME:
			return "CONSECUTIVE_FRAME"
		elif self.type == self.Type.FLOW_CONTROL:
			return "FLOW_CONTROL"
		else:
			return "Reserved"

class TransportLayer:
	"""
	The IsoTP transport layer implementation

	:param rxfn: Function to be called by the transport layer to read the CAN layer. Must return a :class:`isotp.protocol.CanMessage<isotp.protocol.CanMessage>` or None if no message has been received.
	:type rxfn: Callable

	:param txfn: Function to be called by the transport layer to send a message on the CAN layer. This function should receive a :class:`isotp.protocol.CanMessage<isotp.protocol.CanMessage>`
	:type txfn: Callable

	:param address: The address information of CAN messages. Includes the addressing mode, txid/rxid, source/target address and address extension. See :class:`isotp.Address<isotp.Address>` for more details.
	:type address: isotp.Address

	:param error_handler: A function to be called when an error has been detected. An :class:`isotp.protocol.IsoTpError<isotp.protocol.IsoTpError>` (inheriting Exception class) will be given as sole parameter. See the :ref:`Error section<Errors>`
	:type error_handler: Callable

	:param params: List of parameters for the transport layer
	:type params: dict

	"""

	LOGGER_NAME = 'isotp'

	class Params:
		__slots__ = 'stmin', 'blocksize', 'squash_stmin_requirement', 'rx_flowcontrol_timeout', 'rx_consecutive_frame_timeout', 'tx_padding', 'wftmax', 'll_data_length'

		def __init__(self):
			self.stmin 							=  0
			self.blocksize 						=  8
			self.squash_stmin_requirement 		=  False
			self.rx_flowcontrol_timeout 		=  1000
			self.rx_consecutive_frame_timeout 	=  1000
			self.tx_padding 					=  None
			self.wftmax						    = 0
			self.ll_data_length					= 8

		def set(self, key, val):
			setattr(self, key, val)
			self.validate()


		def validate(self):
			if not isinstance(self.rx_flowcontrol_timeout, int):
				raise ValueError('rx_flowcontrol_timeout must be an integer')

			if self.rx_flowcontrol_timeout < 0:
				raise ValueError('rx_flowcontrol_timeout must be positive integer')

			if not isinstance(self.rx_consecutive_frame_timeout, int):
				raise ValueError('rx_consecutive_frame_timeout must be an integer')

			if self.rx_consecutive_frame_timeout < 0:
				raise ValueError('rx_consecutive_frame_timeout must be positive integer')
			
			if self.tx_padding is not None:
				if not isinstance(self.tx_padding, int):
					raise ValueError('tx_padding must be an integer')
				
				if self.tx_padding < 0 or self.tx_padding > 0xFF:
					raise ValueError('tx_padding must be an integer between 0x00 and 0xFF')

			if not isinstance(self.stmin, int):
				raise ValueError('stmin must be an integer')

			if self.stmin < 0 or self.stmin > 0xFF:
				raise ValueError('stmin must be positive integer between 0x00 and 0xFF')

			if not isinstance(self.blocksize, int):
				raise ValueError('blocksize must be an integer')

			if self.blocksize < 0 or self.blocksize > 0xFF:
				raise ValueError('blocksize must be and integer between 0x00 and 0xFF')

			if not isinstance(self.squash_stmin_requirement, bool):
				raise ValueError('squash_stmin_requirement must be a boolean value')

			if not isinstance(self.wftmax, int):
				raise ValueError('wftmax must be an integer')

			if self.wftmax < 0 :
				raise ValueError('wftmax must be and integer equal or greater than 0')

			if not isinstance(self.ll_data_length, int):
				raise ValueError('ll_data_length must be an integer')

			if self.ll_data_length < 4:
				raise ValueError('ll_data_length must be at least 4 bytes')

	class Timer:
		def __init__(self, timeout):
			self.set_timeout(timeout)
			self.start_time = None

		def set_timeout(self, timeout):
			self.timeout = timeout

		def start(self, timeout=None):
			if timeout is not None:
				self.set_timeout(timeout)
			self.start_time = time.time()

		def stop(self):
			self.start_time = None

		def elapsed(self):
			if self.start_time is not None:
				return time.time() - self.start_time
			else:
				return 0

		def is_timed_out(self):
			if self.is_stopped():
				return False
			else:
				return self.elapsed() > self.timeout or self.timeout == 0

		def is_stopped(self):
			return self.start_time == None

	class RxState:
		IDLE = 0
		WAIT_CF = 1

	class TxState:
		IDLE = 0
		WAIT_FC = 1
		TRANSMIT_CF = 2

	def __init__(self, rxfn, txfn, address=None, error_handler=None, params=None):
		self.params = self.Params()
		self.logger = logging.getLogger(self.LOGGER_NAME)

		if params is not None:
			for k in params:
				self.params.set(k, params[k])

		self.remote_blocksize = None	# Block size received in Flow Control message

		self.rxfn = rxfn 	# Function to call to receive a CAN message
		self.txfn = txfn	# Function to call to transmimt a CAN message

		self.set_address(address)

		self.tx_queue = queue.Queue()			# Layer Input queue for IsoTP frame
		self.rx_queue = queue.Queue()			# Layer Output queue for IsoTP frame

		self.rx_state = self.RxState.IDLE		# State of the reception FSM
		self.tx_state = self.TxState.IDLE		# State of the transmission FSM

		self.rx_block_counter = 0
		self.last_seqnum = None					# Consecutive frame Sequence number of previous message	
		self.rx_frame_length = 0				# Length of IsoTP frame being received at the moment
		self.tx_frame_length = 0				# Length of the data that we are sending
		self.last_flow_control_frame = None		# When a FlowControl is received. Put here
		self.tx_block_counter = 0				# Keeps track of how many block we've sent
		self.tx_seqnum = 0						# Keeps track of the actual sequence number whil sending
		self.wft_counter = 0 					# Keeps track of how many wait frame we've received

		self.pending_flow_control_tx = False	# Flag indicating that we need to transmist a flow control message. Set by Rx Process, Cleared by Tx Process
		self.empty_rx_buffer()
		self.empty_tx_buffer()

		self.timer_tx_stmin = self.Timer(timeout = 0)
		self.timer_rx_fc  	= self.Timer(timeout = float(self.params.rx_flowcontrol_timeout) / 1000)
		self.timer_rx_cf 	= self.Timer(timeout = float(self.params.rx_consecutive_frame_timeout) / 1000)

		self.error_handler = error_handler


	
	def send(self, data, target_address_type=isotp.address.TargetAddressType.Physical):
		"""
		Enqueue an IsoTP frame to be sent over CAN network

		:param data: The data to be sent
		:type data: bytearray

		:param target_address_type: Optional parameter that can be Physical (0) for 1-to-1 communication or Functional (1) for 1-to-n. See :class:`isotp.TargetAddressType<isotp.TargetAddressType>`
		:type target_address_type: int

		:raises ValueError: Input parameter is not a bytearray or not convertible to bytearray
		:raises RuntimeError: Transmit queue is full
		"""	
		if not isinstance(data, bytearray):
			try:
				data = bytearray(data)
			except:
				raise ValueError('data must be an IsoTpFrame or bytearray')

		if self.tx_queue.full():
			raise RuntimeError('Transmit queue is full')

		if target_address_type == isotp.address.TargetAddressType.Functional:
			if len(data) > 7-len(self.address.tx_payload_prefix):
				raise ValueError('Cannot send multipacket frame with Functional TargetAddressType')

		self.tx_queue.put( {'data':data, 'target_address_type':target_address_type})	# farme is always an IsoTPFrame here

	# Receive an IsoTP frame. Output of the layer
	def recv(self):
		"""
		Dequeue an IsoTP frame from the reception queue if available.

		:return: The next available IsoTP frame
		:rtype: bytearray or None
		"""	
		if not self.rx_queue.empty():
			return self.rx_queue.get()

	def available(self):
		"""
		Returns ``True`` if an IsoTP frame is awaiting in the reception ``queue``. False otherwise
		"""	
		return not self.rx_queue.empty()

	def transmitting(self):
		"""
		Returns ``True`` if an IsoTP frame is being transmitted. ``False`` otherwise
		"""	
		return not self.tx_queue.empty() or self.tx_state != self.TxState.IDLE

	def process(self):
		"""
		Function to be called periodically, as fast as possible. 
		This function is non-blocking.
		"""	
		msg = True
		while msg is not None:
			msg = self.rxfn()
			if msg is not None:
				self.logger.debug("Receiving : <%03X> %s" % (msg.arbitration_id, binascii.hexlify(msg.data)))
				self.process_rx(msg)

		msg = True
		while msg is not None:
			msg = self.process_tx()
			if msg is not None:
				self.logger.debug("Sending : <%03X> %s" % (msg.arbitration_id, binascii.hexlify(msg.data)))
				self.txfn(msg)

	def process_rx(self, msg):
		
		if not self.address.is_for_me(msg):
			return 

		# Decoding of message into PDU
		try:
			pdu = PDU(msg, start_of_data=self.address.rx_prefix_size, datalen=self.params.ll_data_length)
		except Exception as e:
			self.trigger_error(isotp.errors.InvalidCanDataError("Received invalid CAN frame. %s" % (str(e))))
			self.stop_receiving()
			return

		# Check timeout first
		if self.timer_rx_cf.is_timed_out():
			self.trigger_error(isotp.errors.ConsecutiveFrameTimeoutError("Reception of CONSECUTIVE_FRAME timed out."))
			self.stop_receiving()

		# Process Flow Control message
		if pdu.type == PDU.Type.FLOW_CONTROL:
			self.last_flow_control_frame = pdu 	 # Given to process_tx method. Queue of 1 message depth

			if self.rx_state == self.RxState.WAIT_CF:
				if pdu.flow_status == PDU.FlowStatus.Wait or pdu.flow_status == PDU.FlowStatus.ContinueToSend:
					self.start_rx_cf_timer()
			return # Nothing else to be done with FlowControl. Return and wait for next message

		# Process the state machine
		if self.rx_state == self.RxState.IDLE:
			self.rx_frame_length = 0
			self.timer_rx_cf.stop()

			if pdu.type == PDU.Type.SINGLE_FRAME:
				if pdu.data is not None:
					self.rx_queue.put(copy(pdu.data))

			elif pdu.type == PDU.Type.FIRST_FRAME:
				self.start_reception_after_first_frame(pdu)
			elif pdu.type == PDU.Type.CONSECUTIVE_FRAME:
				self.trigger_error(isotp.errors.UnexpectedConsecutiveFrameError('Received a ConsecutiveFrame while reception was idle. Ignoring'))
				

		elif self.rx_state == self.RxState.WAIT_CF:
			if pdu.type == PDU.Type.SINGLE_FRAME:
				if pdu.data is not None:
					self.rx_queue.put(copy(pdu.data))
					self.rx_state = self.RxState.IDLE
					self.trigger_error(isotp.errors.ReceptionInterruptedWithSingleFrameError('Reception of IsoTP frame interrupted with a new SingleFrame'))

			elif pdu.type == PDU.Type.FIRST_FRAME:
				self.start_reception_after_first_frame(pdu)
				self.trigger_error(isotp.errors.ReceptionInterruptedWithFirstFrameError('Reception of IsoTP frame interrupted with a new FirstFrame'))

			elif pdu.type == PDU.Type.CONSECUTIVE_FRAME:
				self.start_rx_cf_timer() 	# Received a CF message. Restart counter. Timeout handled above.

				expected_seqnum = (self.last_seqnum +1) & 0xF
				if pdu.seqnum == expected_seqnum:
					self.last_seqnum = pdu.seqnum
					bytes_to_receive = (self.rx_frame_length - len(self.rx_buffer) )
					self.append_rx_data(pdu.data[:bytes_to_receive])	# Python handle overflow
					if len(self.rx_buffer) >= self.rx_frame_length:
						self.rx_queue.put(copy(self.rx_buffer))			# Data complete
						self.stop_receiving() 							# Go back to IDLE. Reset all variables and timers.
					else:
						self.rx_block_counter += 1
						if self.rx_block_counter % self.params.blocksize == 0:
							self.request_tx_flowcontrol()  	 # Sets a flag to 1. process_tx will send it for use.
							self.timer_rx_cf.stop() 		 # Deactivate that timer while we wait for flow control
				else:
					self.stop_receiving()
					self.trigger_error(isotp.errors.WrongSequenceNumberError('Received a ConsecutiveFrame with wrong SequenceNumber. Expecting 0x%X, Received 0x%X' % (expected_seqnum, pdu.seqnum)))

	def process_tx(self):
		output_msg = None 	 # Value outputed.  If None, no subsequent call to process_tx will be done.

		# Sends flow control if process_rx requested it
		if self.pending_flow_control_tx:
			self.pending_flow_control_tx = False
			return self.make_flow_control(flow_status=PDU.FlowStatus.ContinueToSend);	# Overflow is not possible. No need to wait.

		# Handle flow control reception
		flow_control_frame = self.last_flow_control_frame	# Reads the last message received and clears it. (Dequeue message)
		self.last_flow_control_frame = None

		if flow_control_frame is not None:
			if flow_control_frame.flow_status == PDU.FlowStatus.Overflow: 	# Needs to stop sending. 
				self.stop_sending()
				return

			if self.tx_state == self.TxState.IDLE:
				self.trigger_error(isotp.errors.UnexpectedFlowControlError('Received a FlowControl message while transmission was Idle. Ignoring'))
			else:
				if flow_control_frame.flow_status == PDU.FlowStatus.Wait:
					if self.params.wftmax == 0:
						self.trigger_error(isotp.errors.UnsuportedWaitFrameError('Received a FlowControl requesting to wait, but fwtmax is set to 0'))
					elif self.wft_counter >= self.params.wftmax:
						self.trigger_error(isotp.errors.MaximumWaitFrameReachedError('Received %d wait frame which is the maximum set in params.wftmax' % (self.wft_counter)))
						self.stop_sending()
					else:
						self.wft_counter += 1
						if self.tx_state in [self.TxState.WAIT_FC, self.TxState.TRANSMIT_CF]:
							self.tx_state = self.TxState.WAIT_FC
							self.start_rx_fc_timer()
				
				elif flow_control_frame.flow_status == PDU.FlowStatus.ContinueToSend and not self.timer_rx_fc.is_timed_out():
					self.wft_counter = 0
					self.timer_rx_fc.stop()
					self.timer_tx_stmin.set_timeout(flow_control_frame.stmin_sec)
					self.remote_blocksize = flow_control_frame.blocksize

					if self.tx_state == self.TxState.WAIT_FC:
						self.tx_block_counter = 0
						self.timer_tx_stmin.start()
					elif self.tx_state == self.TxState.TRANSMIT_CF:
						pass

					self.tx_state = self.TxState.TRANSMIT_CF

		# ======= Timeouts ======
		if self.timer_rx_fc.is_timed_out():
			self.trigger_error(isotp.errors.FlowControlTimeoutError('Reception of FlowControl timed out. Stopping transmission'))
			self.stop_sending()


		# ======= FSM ======
		
		# Check this first as we may have another isotp frame to send and we need to handle it right away without waiting for next "process()" call
		if self.tx_state != self.TxState.IDLE and len(self.tx_buffer) == 0:
			self.stop_sending()	

		if self.tx_state == self.TxState.IDLE:
			read_tx_queue = True	# Read until we get non-empty frame to send
			while read_tx_queue:
				read_tx_queue = False
				if not self.tx_queue.empty():
					popped_object = self.tx_queue.get()
					if len(popped_object['data']) == 0:
						read_tx_queue = True	# Read another frame from tx_queue
					else:
						self.tx_buffer = bytearray(popped_object['data'])

						if len(self.tx_buffer) <= self.params.ll_data_length-1-len(self.address.tx_payload_prefix):	# Single frame
							msg_data 		= self.address.tx_payload_prefix + bytearray([0x0 | len(self.tx_buffer)]) + self.tx_buffer
							arbitration_id 	= self.address.get_tx_arbitraton_id(popped_object['target_address_type'])
							output_msg		= self.make_tx_msg(arbitration_id, msg_data)
						else:							# Multi frame
							data_length = self.params.ll_data_length-2-len(self.address.tx_payload_prefix)
							self.tx_frame_length = len(self.tx_buffer)
							msg_data 		= self.address.tx_payload_prefix + bytearray([0x10|((self.tx_frame_length >> 8) & 0xF), self.tx_frame_length&0xFF]) + self.tx_buffer[:data_length]
							arbitration_id 	= self.address.get_tx_arbitraton_id()
							output_msg 		= self.make_tx_msg(arbitration_id, msg_data)
							self.tx_buffer 	= self.tx_buffer[data_length:]
							self.tx_state 	= self.TxState.WAIT_FC
							self.tx_seqnum 	= 1
							self.start_rx_fc_timer()

		elif self.tx_state == self.TxState.WAIT_FC:
			pass # Nothing to do. Flow control will make the FSM switch state by calling init_tx_consecutive_frame

		elif self.tx_state == self.TxState.TRANSMIT_CF:
			if self.timer_tx_stmin.is_timed_out() or self.params.squash_stmin_requirement:
				data_length = self.params.ll_data_length-1-len(self.address.tx_payload_prefix)
				msg_data = self.address.tx_payload_prefix + bytearray([0x20 | self.tx_seqnum]) + self.tx_buffer[:data_length]
				arbitration_id 	= self.address.get_tx_arbitraton_id()
				output_msg = self.make_tx_msg(arbitration_id, msg_data)
				self.tx_buffer = self.tx_buffer[data_length:]
				self.tx_seqnum = (self.tx_seqnum + 1 ) & 0xF
				self.timer_tx_stmin.start()
				self.tx_block_counter+=1

			if self.remote_blocksize != 0 and self.tx_block_counter >= self.remote_blocksize:
				self.tx_state = self.TxState.WAIT_FC
				self.start_rx_fc_timer()

		return output_msg

	def set_address(self, address):
		"""
		Sets the layer :class:`Address<isotp.Address>`. Can be set after initialization if needed.
		"""

		if not isinstance(address, isotp.address.Address):
			raise ValueError('address must be a valid Address instance')

		self.address = address

		if self.address.txid is not None and (self.address.txid > 0x7F4 and self.address.txid < 0x7F6 or self.address.txid > 0x7FA and self.address.txid < 0x7FB) :
			self.logger.warning('Used txid overlaps the range of ID reserved by ISO-15765 (0x7F4-0x7F6 and 0x7FA-0x7FB)')

		if self.address.rxid is not None and (self.address.rxid > 0x7F4 and self.address.rxid < 0x7F6 or self.address.rxid > 0x7FA and self.address.rxid < 0x7FB) :
			self.logger.warning('Used rxid overlaps the range of ID reserved by ISO-15765 (0x7F4-0x7F6 and 0x7FA-0x7FB)')
			
	def pad_message_data(self, msg_data):
		if len(msg_data) < self.params.ll_data_length and self.params.tx_padding is not None:
			msg_data.extend(bytearray([self.params.tx_padding & 0xFF] * (self.params.ll_data_length-len(msg_data))))
		
	def empty_rx_buffer(self):
		self.rx_buffer = bytearray()

	def empty_tx_buffer(self):
		self.tx_buffer = bytearray()

	def start_rx_fc_timer(self):
		self.timer_rx_fc  	= self.Timer(timeout = float(self.params.rx_flowcontrol_timeout) / 1000)
		self.timer_rx_fc.start()

	def start_rx_cf_timer(self):
		self.timer_rx_cf 	= self.Timer(timeout = float(self.params.rx_consecutive_frame_timeout) / 1000)
		self.timer_rx_cf.start()

	def append_rx_data(self, data):
		self.rx_buffer.extend(data)

	def request_tx_flowcontrol(self):
		self.pending_flow_control_tx = True

	def stop_sending_flow_control(self):
		self.pending_flow_control_tx = False
		self.last_flow_control_frame = None

	def make_tx_msg(self, arbitration_id, data):
		self.pad_message_data(data)
		return CanMessage(arbitration_id = arbitration_id, dlc=len(data), data=data, extended_id=self.address.is_29bits)

	def make_flow_control(self, flow_status=PDU.FlowStatus.ContinueToSend, blocksize=None, stmin=None):
		if blocksize is None:
			blocksize = self.params.blocksize

		if stmin is None:
			stmin = self.params.stmin
		data = PDU.craft_flow_control_data(flow_status, blocksize, stmin)

		return self.make_tx_msg(self.address.get_tx_arbitraton_id(), self.address.tx_payload_prefix + data)

	def request_wait_flow_control(self):
		self.must_wait_for_flow_control = True

	def stop_sending(self):
		self.empty_tx_buffer()
		self.tx_state = self.TxState.IDLE
		self.tx_frame_length = 0
		self.timer_rx_fc.stop()
		self.timer_tx_stmin.stop()
		self.remote_blocksize = None
		self.tx_block_counter = 0
		self.tx_seqnum = 0
		self.wft_counter = 0

	def stop_receiving(self):
		self.rx_state = self.RxState.IDLE
		self.empty_rx_buffer()
		self.stop_sending_flow_control()
		self.timer_rx_cf.stop()

	# Init the reception of a multi-pdu frame. 
	def start_reception_after_first_frame(self, frame):
		self.last_seqnum = 0
		self.rx_block_counter = 0
		self.empty_rx_buffer()
		self.rx_frame_length = frame.length
		self.rx_state = self.RxState.WAIT_CF
		self.append_rx_data(frame.data)
		self.request_tx_flowcontrol()
		self.start_rx_cf_timer()

	def trigger_error(self, error):
		if self.error_handler is not None:
			if hasattr(self.error_handler, '__call__') and isinstance(error, isotp.errors.IsoTpError):
				self.error_handler(error)
			else:
				self.logger.warning('Given error handler is not a callable object.')

		self.logger.warning(str(error))

	# Clears everything within the layer.
	def reset(self):
		"""
		Reset the layer: Empty all buffers, set the internal state machines to Idle
		"""
		while not self.tx_queue.empty():
			self.tx_queue.get()

		while not self.rx_queue.empty():
			self.rx_queue.get()

		self.stop_sending()
		self.stop_receiving()

	# Gives a time to pass to time.sleep() based on the state of the FSM. Avoid using too much CPU
	def sleep_time(self):
		"""
		Returns a value in seconds that can be passed to ``time.sleep()`` when the stack is processed in a different thread.

		The value will change according to the internal state machine state, sleeping longer while idle and shorter when active.
		"""
		timings = {
			(self.RxState.IDLE, self.TxState.IDLE) 		: 0.05,
			(self.RxState.IDLE, self.TxState.WAIT_FC) 	: 0.01,
		}

		key = (self.rx_state, self.tx_state)
		if key in timings:
			return timings[key]
		else:
			return 0.001

class CanStack(TransportLayer):
	"""
	The IsoTP transport using `python-can <https://python-can.readthedocs.io>`_ as CAN layer. python-can must be installed in order to use this class.
	All parameters except the ``bus`` parameter will be given to the :class:`TransportLayer<isotp.TransportLayer>` constructor

	:param bus: A python-can bus object implementing ``recv`` and ``send``
	:type bus: BusABC

	:param address: The address information of CAN messages. Includes the addressing mode, txid/rxid, source/target address and address extension. See :class:`isotp.Address<isotp.Address>` for more details.
	:type address: isotp.Address

	:param error_handler: A function to be called when an error has been detected. An :class:`isotp.protocol.IsoTpError<isotp.protocol.IsoTpError>` (inheriting Exception class) will be given as sole parameter
	:type error_handler: Callable

	:param params: List of parameters for the transport layer
	:type params: dict

	"""

	def tx_canbus(self, msg):
		self.bus.send(can.Message(arbitration_id=msg.arbitration_id, data = msg.data, extended_id=msg.is_extended_id))

	def rx_canbus(self):
		msg = self.bus.recv(0)
		if msg is not None:
			return CanMessage(arbitration_id=msg.arbitration_id, data=msg.data, extended_id=msg.is_extended_id)

	def __init__(self, bus, *args, **kwargs):
		global can
		import can
		self.set_bus(bus)
		TransportLayer.__init__(self, rxfn=self.rx_canbus, txfn=self.tx_canbus, *args, **kwargs)

	def set_bus(self, bus):
		if not isinstance(bus, can.BusABC):
			raise ValueError('bus must be a python-can BusABC object')
		self.bus=bus



import unittest
import isotp
import time
from . import unittest_logging

Message = isotp.CanMessage

# Make sure that our Timer class used for timeouts is working OK
class testTimer(unittest.TestCase):
    def test_timer(self):
        timeout = 0.2
        t = isotp.TransportLayer.Timer(timeout=timeout)
        self.assertFalse(t.is_timed_out())
        self.assertEqual(t.elapsed(), 0)
        t.start()
        self.assertFalse(t.is_timed_out())
        time.sleep(timeout+0.01)
        self.assertTrue(t.elapsed() > timeout)
        self.assertTrue(t.is_timed_out)
        t.stop()
        self.assertFalse(t.is_timed_out())
        self.assertEqual(t.elapsed(), 0)
        t.start()
        self.assertFalse(t.is_timed_out())

# Here we check that we decode properly ecah type of frame
class TestPDUDecoding(unittest.TestCase):

    def make_pdu(self, data, start_of_data=0, datalen=8):
        return isotp.protocol.PDU(Message(data=bytearray(data)),start_of_data=start_of_data, datalen=datalen)

    def test_decode_single_frame(self):
        with self.assertRaises(ValueError):
            self.make_pdu([])

        with self.assertRaises(ValueError):
            self.make_pdu([0])

        for i in range(8,0xF):
            with self.assertRaises(ValueError):
                self.make_pdu([i])

        with self.assertRaises(ValueError):
            self.make_pdu([0x01])

        frame = self.make_pdu([0x01, 0xAA])
        self.assertEqual(frame.type, isotp.protocol.PDU.Type.SINGLE_FRAME)
        self.assertEqual(frame.data, bytearray([0xAA]))
        self.assertEqual(frame.length, len(frame.data))

        with self.assertRaises(ValueError):
            self.make_pdu([0x02, 0x11])

        frame = self.make_pdu([0x02, 0x11, 0x22])
        self.assertEqual(frame.type, isotp.protocol.PDU.Type.SINGLE_FRAME)
        self.assertEqual(frame.data, bytearray([0x11, 0x22]))
        self.assertEqual(frame.length, len(frame.data))

        frame = self.make_pdu([0x02, 0x11, 0x22, 0x33, 0x44])
        self.assertEqual(frame.type, isotp.protocol.PDU.Type.SINGLE_FRAME)
        self.assertEqual(frame.data, bytearray([0x11, 0x22]))
        self.assertEqual(frame.length, len(frame.data))

        frame = self.make_pdu([0x07, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77])
        self.assertEqual(frame.type, isotp.protocol.PDU.Type.SINGLE_FRAME)
        self.assertEqual(frame.data, bytearray([0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77]))
        self.assertEqual(frame.length, len(frame.data))

        with self.assertRaises(ValueError):
            self.make_pdu([0x08,  0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88])

        with self.assertRaises(ValueError):
            self.make_pdu([0x05,  0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88], datalen=5)

        with self.assertRaises(ValueError):
            self.make_pdu([0x01,  0x07, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77], start_of_data = 1, datalen=8)

        with self.assertRaises(ValueError):
            self.make_pdu([0x01,  0x06, 0x22, 0x33, 0x44, 0x55, 0x66], start_of_data = 1, datalen=7)


    def test_decode_first_frame(self):
        with self.assertRaises(ValueError): # Empty payload
            self.make_pdu([])

        with self.assertRaises(ValueError): # Incomplete length
            self.make_pdu([0x10])

        with self.assertRaises(ValueError): # Incomplete length
            self.make_pdu([0x1F])

        with self.assertRaises(ValueError): #Missing data
            self.make_pdu([0x10, 0x02])

        with self.assertRaises(ValueError): # Missing data byte
            self.make_pdu([0x10, 0x02, 0x11])

        frame = self.make_pdu([0x10, 0x02, 0x11, 0x22])
        self.assertEqual(frame.type, isotp.protocol.PDU.Type.FIRST_FRAME)
        self.assertEqual(frame.data, bytearray([0x11, 0x22]))
        self.assertEqual(frame.length, 2)

        frame = self.make_pdu([0x10, 0x02, 0x11, 0x22, 0x33])
        self.assertEqual(frame.type, isotp.protocol.PDU.Type.FIRST_FRAME)
        self.assertEqual(frame.data, bytearray([0x11, 0x22]))
        self.assertEqual(frame.length, 2)

        frame = self.make_pdu([0x10, 0x06, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66])
        self.assertEqual(frame.type, isotp.protocol.PDU.Type.FIRST_FRAME)
        self.assertEqual(frame.data, bytearray([0x11, 0x22, 0x33, 0x44, 0x55, 0x66]))
        self.assertEqual(frame.length, 6)

        with self.assertRaises(ValueError): # Missing data byte
            frame = self.make_pdu([0x10, 0x0A, 0x11, 0x22, 0x33, 0x44, 0x55])

        with self.assertRaises(ValueError): # Missing data byte
            frame = self.make_pdu([0x01, 0x10, 0x0A, 0x11, 0x22, 0x33, 0x44], start_of_data=1)

        with self.assertRaises(ValueError): # Missing data byte
            frame = self.make_pdu([0x01, 0x10, 0x0A, 0x11, 0x22, 0x33, 0x44, 0x55], start_of_data=1, datalen=12)

        frame = self.make_pdu([0x10, 0x0A, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66])
        self.assertEqual(frame.type, isotp.protocol.PDU.Type.FIRST_FRAME)
        self.assertEqual(frame.data, bytearray([0x11, 0x22, 0x33, 0x44, 0x55, 0x66]))
        self.assertEqual(frame.length, 0xA)

        frame = self.make_pdu([0x1A, 0xBC, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66])
        self.assertEqual(frame.type, isotp.protocol.PDU.Type.FIRST_FRAME)
        self.assertEqual(frame.data, bytearray([0x11, 0x22, 0x33, 0x44, 0x55, 0x66]))
        self.assertEqual(frame.length, 0xABC)

        frame = self.make_pdu([0x00, 0x00, 0x1A, 0xBC, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88, 0x99], start_of_data=2, datalen=12)
        self.assertEqual(frame.type, isotp.protocol.PDU.Type.FIRST_FRAME)
        self.assertEqual(frame.data, bytearray([0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88]))
        self.assertEqual(frame.length, 0xABC)

    
    def test_decode_consecutive_frame(self):
        with self.assertRaises(ValueError): # Empty payload
            self.make_pdu([])

        frame = self.make_pdu([0x20])
        self.assertEqual(frame.type, isotp.protocol.PDU.Type.CONSECUTIVE_FRAME)
        self.assertEqual(frame.data, bytearray([]))
        self.assertEqual(frame.seqnum, 0)

        frame = self.make_pdu([0x20, 0x11])
        self.assertEqual(frame.type, isotp.protocol.PDU.Type.CONSECUTIVE_FRAME)
        self.assertEqual(frame.data, bytearray([0x11]))
        self.assertEqual(frame.seqnum, 0)

        frame = self.make_pdu([0x2A, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77])
        self.assertEqual(frame.type, isotp.protocol.PDU.Type.CONSECUTIVE_FRAME)
        self.assertEqual(frame.data, bytearray([0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77]))
        self.assertEqual(frame.seqnum, 0xA)

        frame = self.make_pdu([0x00, 0x2A, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77], start_of_data=1, datalen=6)
        self.assertEqual(frame.type, isotp.protocol.PDU.Type.CONSECUTIVE_FRAME)
        self.assertEqual(frame.data, bytearray([0x11, 0x22, 0x33, 0x44]))
        self.assertEqual(frame.seqnum, 0xA)


    def test_decode_flow_control(self):
        with self.assertRaises(ValueError): # Empty payload
            self.make_pdu([])

        with self.assertRaises(ValueError): # incomplete
            self.make_pdu([0x30])

        with self.assertRaises(ValueError): # incomplete
            self.make_pdu([0x30, 0x00])

        frame = self.make_pdu([0x30, 0x00, 0x00])
        self.assertEqual(frame.type, isotp.protocol.PDU.Type.FLOW_CONTROL)
        self.assertEqual(frame.flow_status, isotp.protocol.PDU.FlowStatus.ContinueToSend)
        self.assertEqual(frame.blocksize, 0)
        self.assertEqual(frame.stmin, 0)
        self.assertEqual(frame.stmin_sec, 0)

        frame = self.make_pdu([0x31, 0x00, 0x00])
        self.assertEqual(frame.type, isotp.protocol.PDU.Type.FLOW_CONTROL)
        self.assertEqual(frame.flow_status, isotp.protocol.PDU.FlowStatus.Wait)
        self.assertEqual(frame.blocksize, 0)
        self.assertEqual(frame.stmin, 0)
        self.assertEqual(frame.stmin_sec, 0)

        frame = self.make_pdu([0x32, 0x00, 0x00])
        self.assertEqual(frame.type, isotp.protocol.PDU.Type.FLOW_CONTROL)
        self.assertEqual(frame.flow_status, isotp.protocol.PDU.FlowStatus.Overflow)
        self.assertEqual(frame.blocksize, 0)
        self.assertEqual(frame.stmin, 0)
        self.assertEqual(frame.stmin_sec, 0)

        frame = self.make_pdu([0xFF, 0xFF, 0x32, 0x01, 0x01], start_of_data=2)
        self.assertEqual(frame.type, isotp.protocol.PDU.Type.FLOW_CONTROL)
        self.assertEqual(frame.flow_status, isotp.protocol.PDU.FlowStatus.Overflow)
        self.assertEqual(frame.blocksize, 1)
        self.assertEqual(frame.stmin, 1)
        self.assertEqual(frame.stmin_sec, 1/1000)

        for i in range(3, 0xF): # Reserved Flow status
            with self.assertRaises(ValueError):
                frame = self.make_pdu([0x30 + i, 0x00, 0x00])

        frame = self.make_pdu([0x30, 0xFF, 0x00])
        self.assertEqual(frame.type, isotp.protocol.PDU.Type.FLOW_CONTROL)
        self.assertEqual(frame.flow_status, isotp.protocol.PDU.FlowStatus.ContinueToSend)
        self.assertEqual(frame.blocksize, 0xFF)
        self.assertEqual(frame.stmin, 0)
        self.assertEqual(frame.stmin_sec, 0)

        for i in range(0,0x7F):     # Millisecs
            frame = self.make_pdu([0x30, 0x00, i])
            self.assertEqual(frame.type, isotp.protocol.PDU.Type.FLOW_CONTROL)
            self.assertEqual(frame.flow_status, isotp.protocol.PDU.FlowStatus.ContinueToSend)
            self.assertEqual(frame.blocksize, 0)
            self.assertEqual(frame.stmin, i)
            self.assertEqual(frame.stmin_sec, i/1000)

        for i in range(0xF1, 0xF9): # Microsecs
            frame = self.make_pdu([0x30, 0x00, i])
            self.assertEqual(frame.type, isotp.protocol.PDU.Type.FLOW_CONTROL)
            self.assertEqual(frame.flow_status, isotp.protocol.PDU.FlowStatus.ContinueToSend)
            self.assertEqual(frame.blocksize, 0)
            self.assertEqual(frame.stmin, i)
            self.assertEqual(frame.stmin_sec, (i - 0xF0)/10000)

        for i in range(0x80, 0xF1):     # Reserved StMin
            with self.assertRaises(ValueError):
                frame = self.make_pdu([0x30, 0x00, i])

        for i in range(0xFA, 0x100):    # Reserved StMin
            with self.assertRaises(ValueError):
                frame = self.make_pdu([0x30, 0x00, i])


import unittest
import time
import isotp
from test.ThreadableTest import ThreadableTest
from . import unittest_logging
from . import tools
import math


@unittest.skipIf(tools.check_isotp_socket_possible() == False, 'Cannot test stack against IsoTP socket. %s' % tools.isotp_socket_impossible_reason())
class TestLayerAgainstSocket(ThreadableTest):
    
    def __init__(self, *args, **kwargs):
        ThreadableTest.__init__(self, *args, **kwargs)
        if not hasattr(self.__class__, '_next_id'):
            self.__class__._next_id=1

        (self.stack_txid, self.stack_rxid) = tools.get_next_can_id_pair()

        self.transmission_complete = False
        self.reception_complete = False
        self.socket_ready=False

        global can
        import can

    def setUp(self):
        self.socket_list = []

    def make_socket(self, *args, **kwargs):
        s = isotp.socket(*args, **kwargs)
        self.socket_list.append(s)
        return s

    def clientSetUp(self):
        self.bus = can.interface.Bus(tools.get_test_interface(), bustype='socketcan')
        address = isotp.Address(isotp.AddressingMode.Normal_11bits, txid=self.stack_txid, rxid=self.stack_rxid)
        self.stack = isotp.CanStack(bus=self.bus, address=address)

    def tearDown(self):
        for socket in self.socket_list:
            socket.close()
        self.stack.reset()
        self.bus.shutdown()

    def process_stack_receive(self, timeout=1):
        t1 = time.time()
        self.reception_complete = False
        while time.time() - t1 < timeout:
            self.stack.process()
            if self.stack.available():
                break
            time.sleep(self.stack.sleep_time())
        self.reception_complete = True
        return self.stack.recv()

    def process_stack_send(self, timeout=1):
        self.transmission_complete = False
        t1 = time.time()
        while time.time() - t1 < timeout:
            self.stack.process()
            if not self.stack.transmitting():
                break
            time.sleep(self.stack.sleep_time())
        self.transmission_complete = True

    def wait_transmission_complete(self, timeout=1):
        t1 = time.time()
        while time.time() - t1 < timeout and self.transmission_complete == False:
            time.sleep(0.05)

    def wait_reception_complete(self, timeout=1):
        t1 = time.time()
        while time.time() - t1 < timeout and self.reception_complete == False:
            time.sleep(0.05)

    def wait_socket_ready(self, timeout=1):
        t1 = time.time()
        while time.time() - t1 < timeout and self.socket_ready == False:
            time.sleep(0.05)
        


    # ========= Test cases ======
    def test_receive(self):
        s = self.make_socket()
        s.bind(tools.get_test_interface(), txid=self.stack_rxid, rxid=self.stack_txid)
        self.socket_ready = True
        s.send(b'a'*100)
        self.wait_reception_complete()

    def _test_receive(self):
        self.wait_socket_ready()
        frame = self.process_stack_receive()
        self.assertEqual(frame, b'a'*100)

    def test_transmit(self):
        s = self.make_socket()
        s.bind(tools.get_test_interface(), txid=self.stack_rxid, rxid=self.stack_txid)
        self.socket_ready = True

        self.wait_transmission_complete(1)
        frame = s.recv()
        self.assertEqual(frame, b'b'*100)

    def _test_transmit(self):
        self.wait_socket_ready()
        self.stack.send(b'b'*100)
        self.process_stack_send()
        
    def test_transmit_long_stmin(self):
        s = self.make_socket()
        s.set_fc_opts(stmin=100)
        s.bind(tools.get_test_interface(), txid=self.stack_rxid, rxid=self.stack_txid)
        self.socket_ready = True
        self.wait_transmission_complete(5)
        frame = s.recv()
        self.assertEqual(frame, b'b'*150)

    def _test_transmit_long_stmin(self):
        self.wait_socket_ready()
        payload = b'b'*150
        ncf = math.ceil(max(len(payload)-6,0)/7)
        expected_time = ncf * 0.1
        self.stack.send(payload)
        t1 = time.time()
        self.process_stack_send(timeout=2*expected_time*0.95)
        diff = time.time() - t1
        self.assertGreater(diff, expected_time)

    def test_receive_long_stmin(self):
        s = self.make_socket()
        s.bind(tools.get_test_interface(), txid=self.stack_rxid, rxid=self.stack_txid)
        self.socket_ready = True
        s.send(b'a'*150)
        self.wait_reception_complete(timeout=5)

    def _test_receive_long_stmin(self):
        self.wait_socket_ready()
        payload = b'a'*150
        ncf = math.ceil(max(len(payload)-6,0)/7)
        expected_time = ncf * 0.1
        t1 = time.time()
        self.stack.params.set('stmin', 100)
        frame = self.process_stack_receive(timeout=2*expected_time)
        self.assertEqual(frame, payload)
        diff = time.time() - t1
        self.assertGreater(diff, expected_time*0.95)

    def test_receive_extended_29bits(self):
        s = self.make_socket()
        addr = isotp.Address(isotp.AddressingMode.Extended_29bits, txid=self.stack_rxid, rxid=self.stack_txid, source_address=0x88, target_address=0x99)
        s.bind(tools.get_test_interface(),addr)
        self.socket_ready = True
        s.send(b'a'*100)
        self.wait_reception_complete()

    def _test_receive_extended_29bits(self):
        self.wait_socket_ready()
        addr = isotp.Address(isotp.AddressingMode.Extended_29bits, txid=self.stack_txid, rxid=self.stack_rxid, source_address=0x99, target_address=0x88)
        self.stack.set_address(addr)
        frame = self.process_stack_receive()
        self.assertEqual(frame, b'a'*100)

    def test_transmit_extended_29bits(self):
        s = self.make_socket()
        addr = isotp.Address(isotp.AddressingMode.Extended_29bits, txid=self.stack_rxid, rxid=self.stack_txid, source_address=0x88, target_address=0x99)
        s.bind(tools.get_test_interface(), addr)
        self.socket_ready = True

        self.wait_transmission_complete(1)
        frame = s.recv()
        self.assertEqual(frame, b'b'*100)

    def _test_transmit_extended_29bits(self):
        self.wait_socket_ready()
        addr = isotp.Address(isotp.AddressingMode.Extended_29bits, txid=self.stack_txid, rxid=self.stack_rxid, source_address=0x99, target_address=0x88)
        self.stack.set_address(addr)
        self.stack.send(b'b'*100)
        self.process_stack_send()

    def test_receive_mixed_29bits(self):
        s = self.make_socket()
        addr = isotp.Address(isotp.AddressingMode.Mixed_29bits, source_address=0x88, target_address=0x99, address_extension=0xDD)
        s.bind(tools.get_test_interface(),addr)
        self.socket_ready = True
        s.send(b'c'*100)
        self.wait_reception_complete()

    def _test_receive_mixed_29bits(self):
        self.wait_socket_ready()
        addr = isotp.Address(isotp.AddressingMode.Mixed_29bits, source_address=0x99, target_address=0x88, address_extension=0xDD)
        self.stack.set_address(addr)
        frame = self.process_stack_receive()
        self.assertEqual(frame, b'c'*100)

    def test_transmit_mixed_29bits(self):
        s = self.make_socket()
        addr = isotp.Address(isotp.AddressingMode.Mixed_29bits, source_address=0x88, target_address=0x99, address_extension=0xEE)
        s.bind(tools.get_test_interface(), addr)
        self.socket_ready = True

        self.wait_transmission_complete(1)
        frame = s.recv()
        self.assertEqual(frame, b'd'*100)

    def _test_transmit_mixed_29bits(self):
        self.wait_socket_ready()
        addr = isotp.Address(isotp.AddressingMode.Mixed_29bits, source_address=0x99, target_address=0x88, address_extension=0xEE)
        self.stack.set_address(addr)
        self.stack.send(b'd'*100)
        self.process_stack_send()


import socket as socket_module
import struct
from . import socket
from . import check_support
check_support()

flags = socket.flags

def assert_is_socket(s):
    if not isinstance(s, socket_module.socket):
        raise ValueError("Given value is not a socket.")

SOL_CAN_BASE        = socket_module.SOL_CAN_BASE if hasattr(socket_module, 'SOL_CAN_BASE') else 100
SOL_CAN_ISOTP       =  SOL_CAN_BASE + socket_module.CAN_ISOTP if hasattr(socket_module, 'CAN_ISOTP') else None # Allow importing the module without python 3.7
CAN_ISOTP_OPTS      = 1
CAN_ISOTP_RECV_FC   = 2
CAN_ISOTP_TX_STMIN  = 3
CAN_ISOTP_RX_STMIN  = 4
CAN_ISOTP_LL_OPTS   = 5


class general:
    struct_size = 4+4+1+1+1+1

    def __init__(self):
        self.optflag = None
        self.frame_txtime = None
        self.ext_address = None
        self.txpad = None
        self.rxpad = None
        self.rx_ext_address = None

    
    @classmethod
    def read(cls, s):
        assert_is_socket(s)
        o = cls()
        opt = s.getsockopt(SOL_CAN_ISOTP, CAN_ISOTP_OPTS,cls.struct_size)

        (o.optflag, o.frame_txtime, o.ext_address, o.txpad, o.rxpad, o.rx_ext_address) = struct.unpack("=LLBBBB", opt)
        return o


    @classmethod
    def write(cls, s, optflag=None, frame_txtime=None, ext_address=None, txpad=None, rxpad=None, rx_ext_address=None ):
        assert_is_socket(s)
        o = cls.read(s);

        if optflag != None:
            if not isinstance(optflag, int) or optflag<0 or optflag>0xFFFFFFFF:
                raise ValueError("optflag must be a valid 32 unsigned integer")
            o.optflag = optflag

        if frame_txtime != None:
            if not isinstance(frame_txtime, int) or frame_txtime<0 or frame_txtime>0xFFFFFFFF:
                raise ValueError("frame_txtime must be a valid 32 unsigned integer")
            o.frame_txtime = frame_txtime
            o.optflag |= flags.FORCE_TXSTMIN

        if ext_address != None:
            if not isinstance(ext_address, int) or ext_address<0 or ext_address>0xFF:
                raise ValueError("ext_address must be a an integer between 0 and FF")
            o.ext_address = ext_address
            o.optflag |= flags.EXTEND_ADDR

        if txpad != None:
            if not isinstance(txpad, int) or txpad<0 or txpad>0xFF:
                raise ValueError("txpad must be a an integer between 0 and FF")
            o.txpad = txpad
            o.optflag |= flags.TX_PADDING

        if rxpad != None:
            if not isinstance(rxpad, int) or rxpad<0 or rxpad>0xFF:
                raise ValueError("rxpad must be a an integer between 0 and FF")
            o.rxpad = rxpad
            o.optflag |= flags.RX_PADDING

        if rx_ext_address != None:
            if not isinstance(rx_ext_address, int) or rx_ext_address<0 or rx_ext_address>0xFF:
                raise ValueError("rx_ext_address must be a an integer between 0 and FF")
            o.rx_ext_address = rx_ext_address
            o.optflag |= flags.RX_EXT_ADDR

        opt = struct.pack("=LLBBBB", o.optflag, o.frame_txtime, o.ext_address, o.txpad, o.rxpad, o.rx_ext_address)
        s.setsockopt(SOL_CAN_ISOTP, CAN_ISOTP_OPTS, opt)
        return o

    def __repr__(self):
        optflag_str = '[undefined]' if self.optflag is None else '0x%x' % (self.optflag)
        frame_txtime_str = '[undefined]' if self.frame_txtime is None else '0x%x' % (self.frame_txtime)
        ext_address_str = '[undefined]' if self.ext_address is None else '0x%x' % (self.ext_address)
        txpad_str = '[undefined]' if self.txpad is None else '0x%x' % (self.txpad)
        rxpad_str = '[undefined]' if self.rxpad is None else '0x%x' % (self.rxpad)
        rx_ext_address_str = '[undefined]' if self.rx_ext_address is None else '0x%x' % (self.rx_ext_address)

        return "<OPTS_GENERAL: optflag=%s, frame_txtime=%s, ext_address=%s, txpad=%s, rxpad=%s, rx_ext_address=%s>" % (optflag_str, frame_txtime_str, ext_address_str, txpad_str, rxpad_str, rx_ext_address_str)

class flowcontrol:
    struct_size = 3
    
    def __init__(self):
        self.stmin = None;
        self.bs = None;
        self.wftmax = None;

    @classmethod
    def read(cls, s):
        assert_is_socket(s)
        o = cls()
        opt = s.getsockopt(SOL_CAN_ISOTP, CAN_ISOTP_RECV_FC,cls.struct_size)

        (o.bs, o.stmin, o.wftmax) = struct.unpack("=BBB", opt)
        return o

    @classmethod
    def write(cls, s, bs=None, stmin=None, wftmax=None):
        assert_is_socket(s)
        o = cls.read(s);
        if bs != None:
            if not isinstance(bs, int) or bs<0 or bs>0xFF:
                raise ValueError("bs must be a valid interger between 0 and FF")
            o.bs = bs

        if stmin != None:
            if not isinstance(stmin, int) or stmin<0 or stmin>0xFF:
                raise ValueError("stmin must be a valid interger between 0 and FF")
            o.stmin = stmin

        if wftmax != None:
            if not isinstance(wftmax, int) or wftmax<0 or wftmax>0xFF:
                raise ValueError("wftmax must be a valid interger between 0 and FF")
            o.wftmax = wftmax

        opt = struct.pack("=BBB", o.bs, o.stmin, o.wftmax)
        s.setsockopt(SOL_CAN_ISOTP, CAN_ISOTP_RECV_FC, opt)
        return o

    def __repr__(self):
        bs_str = '[undefined]' if self.bs is None else '0x%x' % (self.bs)
        stmin_str = '[undefined]' if self.stmin is None else '0x%x' % (self.stmin)
        wftmax_str = '[undefined]' if self.wftmax is None else '0x%x' % (self.wftmax)
        return "<OPTS_RECV_FC: bs=%s, stmin=%s, wftmax=%s>" % (bs_str, stmin_str, wftmax_str)

class linklayer:
    struct_size = 3
    
    def __init__(self):
        self.mtu = None;
        self.tx_dl = None;
        self.tx_flags = None;

    @classmethod
    def read(cls, s):
        assert_is_socket(s)
        o = cls()
        opt = s.getsockopt(SOL_CAN_ISOTP, CAN_ISOTP_LL_OPTS,cls.struct_size)

        (o.mtu, o.tx_dl, o.tx_flags) = struct.unpack("=BBB", opt)
        return o

    @classmethod
    def write(cls, s, mtu=None, tx_dl=None, tx_flags=None):
        assert_is_socket(s)
        o = cls.read(s);
        if mtu != None:
            if not isinstance(mtu, int) or mtu<0 or mtu>0xFF:
                raise ValueError("mtu must be a valid interger between 0 and FF")
            o.mtu = mtu

        if tx_dl != None:
            if not isinstance(tx_dl, int) or tx_dl<0 or tx_dl>0xFF:
                raise ValueError("tx_dl must be a valid interger between 0 and FF")
            o.tx_dl = tx_dl

        if tx_flags != None:
            if not isinstance(tx_flags, int) or tx_flags<0 or tx_flags>0xFF:
                raise ValueError("tx_flags must be a valid interger between 0 and FF")
            o.tx_flags = tx_flags

        opt = struct.pack("=BBB", o.mtu, o.tx_dl, o.tx_flags)
        s.setsockopt(SOL_CAN_ISOTP, CAN_ISOTP_LL_OPTS, opt)
        return o

    def __repr__(self):
        mtu_str = '[undefined]' if self.mtu is None else '0x%x' % (self.mtu)
        tx_dl_str = '[undefined]' if self.tx_dl is None else '0x%x' % (self.tx_dl)
        tx_flags_str = '[undefined]' if self.tx_flags is None else '0x%x' % (self.tx_flags)
        return "<OPTS_LL: mtu=%s, tx_dl=%s, tx_flags=%s>" % (mtu_str, tx_dl_str, tx_flags_str)


class Vars:
    SOCKET_IMPOSSIBLE_REASON = ''
    interface_name = 'vcan0'
    ISOTP_SOCKET_POSSIBLE = False
    
    next_id = 10

def check_isotp_socket_possible():
    try:
        Vars.SOCKET_IMPOSSIBLE_REASON = ''
        import isotp
        import can
        s = isotp.socket()
        s.bind(Vars.interface_name, rxid=1, txid=2)
        s.close()
        Vars.ISOTP_SOCKET_POSSIBLE = True
    except Exception as e:
        Vars.SOCKET_IMPOSSIBLE_REASON = str(e)
        Vars.ISOTP_SOCKET_POSSIBLE = False

    return Vars.ISOTP_SOCKET_POSSIBLE

def isotp_socket_impossible_reason():
    return Vars.SOCKET_IMPOSSIBLE_REASON

def get_test_interface():
    return Vars.interface_name

def get_next_can_id_pair():
    pair = (Vars.next_id, Vars.next_id+1)
    Vars.next_id += 2
    return pair